# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
nbd - Network Block Device
"""

import errno
import itertools
import logging
import os
import re
import socket
import struct

from . import ipv6
from . import sockutil
from .units import MiB

# Matcher for NBD Unix URL path.
# nbd:unix:path[:exportname=name]
UNIX_URL_PATH = re.compile(
    r"unix:(?P<path>/[^:]+)(:?:exportname=(?P<name>.*))?")

# Matcher for NBD TCP URL path.
# nbd:host:port[:exportname=name]
TCP_URL_PATH = re.compile(
    r"(?P<host>.+):(?P<port>\d+)(:?:exportname=(?P<name>.*))?")

# Magic numbers.
NBDMAGIC = 0x4e42444d41474943
IHAVEOPT = 0x49484156454F5054
OPTION_REPLY_MAGIC = 0x3e889045565a9
REQUEST_MAGIC = 0x25609513
SIMPLE_REPLY_MAGIC = 0x67446698
STRUCTURED_REPLY_MAGIC = 0x668e33ef

# Flags
FLAG_FIXED_NEWSTYLE = 1
FLAG_NO_ZEROES = 2
FLAG_C_FIXED_NEWSTYLE = 1
FLAG_C_NO_ZEROES = 2

# Transmission flags
FLAG_HAS_FLAGS = (1 << 0)
FLAG_READ_ONLY = (1 << 1)
FLAG_SEND_FLUSH = (1 << 2)
FLAG_SEND_FUA = (1 << 3)
FLAG_ROTATIONAL = (1 << 4)
FLAG_SEND_TRIM = (1 << 5)
FLAG_SEND_WRITE_ZEROES = (1 << 6)
FLAG_SEND_DF = (1 << 7)
FLAG_CAN_MULTI_CONN = (1 << 8)
FLAG_SEND_RESIZE = (1 << 9)
FLAG_SEND_CACHE = (1 << 10)

# Options
OPT_ABORT = 2
OPT_GO = 7
OPT_STRUCTURED_REPLY = 8
OPT_LIST_META_CONTEXT = 9
OPT_SET_META_CONTEXT = 10

# Replies
REP_ACK = 1
REP_INFO = 3
REP_META_CONTEXT = 4

# Structured reply flags
REPLY_FLAG_DONE = (1 << 0)

# Flags for base:allocation meta context.

# This range does not allocate any data on storage. Examples are a hole
# in raw image, or zero cluster in qcow2 image. This is flag is optional
# and may be ommited by a NBD server. Cannot be used to detect
# unallocated areas in qcow2 exposing data from the backing file.
STATE_HOLE = (1 << 0)

# The range is read as zero.
STATE_ZERO = (1 << 1)

# Flags for qemu:dirty-bitmap meta context.
STATE_DIRTY = (1 << 0)

# Adusted to support merging allocation and dirty bits.
EXTENT_DIRTY = (1 << 2)

# Extent does not exist, exposing data from the backing file.
EXTENT_BACKING = (1 << 3)

# Command flags
CMD_FLAG_NO_HOLE = (1 << 1)

# Structured reply types
REPLY_TYPE_NONE = 0
REPLY_TYPE_OFFSET_DATA = 1
REPLY_TYPE_OFFSET_HOLE = 2
REPLY_TYPE_BLOCK_STATUS = 5
REPLY_ERROR_BASE = (1 << 15)
REPLY_TYPE_ERROR = REPLY_ERROR_BASE + 1
REPLY_TYPE_ERROR_OFFSET = REPLY_ERROR_BASE + 2

# NBD_INFO replies
INFO_EXPORT = 0
INFO_BLOCK_SIZE = 3

# Meta context names.
BASE_ALLOCATION = "base:allocation"
QEMU_ALLOCATION_DEPTH = "qemu:allocation-depth"
QEMU_DIRTY_BITMAP = "qemu:dirty-bitmap:"

# Error replies
ERR_BASE = 2**31
REP_ERR_UNSUP = ERR_BASE + 1
REP_ERR_POLICY = ERR_BASE + 2
REP_ERR_INVALID = ERR_BASE + 3
REP_ERR_PLATFORM = ERR_BASE + 4
REP_ERR_TLS_REQD = ERR_BASE + 5
REP_ERR_UNKNOWN = ERR_BASE + 6
REP_ERR_SHUTDOWN = ERR_BASE + 7
REP_ERR_BLOCK_SIZE_REQD = ERR_BASE + 8
REP_ERR_TOO_BIG = ERR_BASE + 9

ERROR_REPLY = {
    REP_ERR_UNSUP: (
        "The option sent by the client is unknown by this server "
        "implementation"),
    REP_ERR_POLICY: (
        "The option sent by the client is known by this server and "
        "syntactically valid, but server-side policy forbids the server to "
        "allow the option"),
    REP_ERR_INVALID: (
        "The option sent by the client is known by this server, but was "
        "determined by the server to be syntactically or semantically "
        "invalid"),
    REP_ERR_PLATFORM: (
        "The option sent by the client is not supported on the platform on "
        "which the server is running"),
    REP_ERR_TLS_REQD: (
        "The server is unwilling to continue negotiation unless TLS is "
        "initiated first"),
    REP_ERR_UNKNOWN: "The requested export is not available",
    REP_ERR_SHUTDOWN: (
        "The server is unwilling to continue negotiation as it is in the "
        "process of being shut down"),
    REP_ERR_BLOCK_SIZE_REQD: (
        "The server is unwilling to enter transmission phase for a given "
        "export unless the client first acknowledges (via "
        "INFO_BLOCK_SIZE) that it will obey non-default block sizing "
        "requirements"),
    REP_ERR_TOO_BIG: "The request or the reply is too large to process",
}

# Mapping from NBD error code in simple or structured reply to system errno.
# https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
# #error-values
REPLY_ERRORS = {
    1: errno.EPERM,
    5: errno.EIO,
    12: errno.ENOMEM,
    22: errno.EINVAL,
    28: errno.ENOSPC,
    75: errno.EOVERFLOW,
    108: errno.ESHUTDOWN,
}

# NBD Option struct
# C: 64 bits, 0x49484156454F5054 (ASCII 'IHAVEOPT')
# C: 32 bits, option
# C: 32 bits, length of option data (unsigned)
OPTION = struct.Struct("!QII")

# Maximum NBD request length (unsigned 32 bit integer).
MAX_LENGTH = 2**32 - 1

# The NBD spec does not define how many extents chunks a server may send in
# REPLY_TYPE_BLOCK_STATUS.  Theoretically a server can retrun one extent per
# byte if the minimum block size is 1 byte. Practically for raw images minimum
# extent size is the file system block size (likely 4 KiB), and for qcow2
# images the cluster size (likely 64 KiB). Since we don't know the image
# format, assume raw image. If a server send more extents than this value we
# fail the connection.
MAX_EXTENTS = MAX_LENGTH // 4096

log = logging.getLogger("nbd")


class Error(Exception):
    fmt = "{self.reason}"

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.fmt.format(self=self)


class ProtocolError(Error):
    """
    Raised when server sent invalid response. Requires termination of the
    connection.
    """


class UnexpectedOptionReply(ProtocolError):
    fmt = ("Unexpected reply {self.reply} for option {self.option}, "
           "expecting reply {self.expected}")

    def __init__(self, reply, option, expected):
        self.reply = reply
        self.option = option
        self.expected = expected


class InvalidLength(ProtocolError):
    fmt = ("Reply {self.reply} with invalid legnth {self.length}, expecting "
           "{self.expected}")

    def __init__(self, reply, length, expected):
        self.reply = reply
        self.length = length
        self.expected = expected


class UnexpectedHandle(ProtocolError):
    fmt = "Unepected handle {self.handle}, expecting {self.expected}"

    def __init__(self, handle, expected):
        self.handle = handle
        self.expected = expected


class OptionError(Error):
    fmt = ("Error negotiating option opt={self.opt} code={self.code} "
           "reason={self.reason}")

    def __init__(self, opt, code, reason):
        self.opt = opt
        self.code = code
        self.reason = reason


class OptionUnsupported(OptionError):
    fmt = "Option {self.option} is not supported: {self.reason}"
    code = REP_ERR_UNSUP

    def __init__(self, option, reason):
        self.option = option
        self.reason = reason


class RequestError(Error):
    """
    Raised when server failed to process a request. The client can continue
    normally with another request or repeat the failing request.
    """


class UnsupportedRequest(RequestError):
    """
    Raised when the server cannot process a request becuase the requested
    operation is not supported for the current connection. The client should
    not send the same request again.
    """


class ReplyError(RequestError):
    """
    Raised when server return an error reply.
    """
    fmt = "{self.message}: [Error {self.code}] {self.reason}"

    def __init__(self, code, message):
        """
        Arguments:
            code (int): NBD reply error code.
            message (str): string suitable for displaying to the user.
        """
        # Message is optional, but qemu-nbd always sends a message.
        if not message:
            message = "Server error"

        self.code = code
        self.message = message.capitalize()
        self.reason = os.strerror(REPLY_ERRORS.get(code, code))


class UnixAddress(sockutil.UnixAddress):
    """
    sockutil.UnixAddress enriched with url() providing nbd URL.
    """

    def url(self, export=None):
        s = "nbd:unix:{}".format(self.path)
        if export:
            s += ":exportname=" + export
        return s


class TCPAddress(sockutil.TCPAddress):
    """
    sockutil.TCPAddress enriched with url() providing nbd URL.
    """
    def url(self, export=None):
        host = ipv6.quote_address(self.host)
        s = "nbd:{}:{}".format(host, self.port)
        if export:
            s += ":exportname=" + export
        return s


def open(url, dirty=False):
    """
    Open parsed NBD URL and return a connected Client instance.
    """
    address, name = _parse_url(url)
    return Client(address, name, dirty=dirty)


def _parse_url(url):
    """
    Parse url and return 2 tuple (address, name), or raise an Error.
    """
    if url.scheme != "nbd":
        raise Error("Unsupported URL scheme: {}".format(url))

    # First try the nice URL notation:
    # nbd://localhost:10809/sda
    # This notiation is less flexible but nicer for humans.
    # See https://qemu.weilnetz.de/doc/qemu-doc.html#disk_005fimages_005fnbd
    if ":" in url.netloc:
        host, port = url.netloc.rsplit(":", 1)
        export = url.path
        # According to NBD spec (and qemu implementation) the / starting the
        # path component of the URL is not considered part of the export name.
        # To create export name with leading /, the path must start with //.
        # https://github.com/NetworkBlockDevice/nbd/blob/master/doc/uri.md
        if export.startswith("/"):
            export = export[1:]
        return TCPAddress(host, int(port)), export

    # Next try to documented NBD URL notation. This notiation is more flexible
    # and can handle export names with leading "/".
    # - nbd:unix:path[:exportname=name]
    # - nbd:host:port[:exportname=name]
    # See https://qemu.weilnetz.de/doc/qemu-doc.html#Device-URL-Syntax
    if url.netloc == "":
        match = UNIX_URL_PATH.match(url.path)
        if match:
            d = match.groupdict()
            return UnixAddress(d["path"]), d["name"]

        match = TCP_URL_PATH.match(url.path)
        if match:
            d = match.groupdict()
            return TCPAddress(d["host"], int(d["port"])), d["name"]

    raise Error("Unsupported URL: {}".format(url))


# Client states

CONNECTING = 0
HANDSHAKE = 1
TRANSMISSION = 2
CLOSED = 3


class Client:

    def __init__(self, address, export_name=None, dirty=False):
        self.address = address
        self.export_name = export_name or ""
        self.dirty = dirty

        log.debug("Connecting address=%r export_name=%r dirty=%r",
                  address, self.export_name, dirty)

        self.export_size = None
        self.transmission_flags = None

        # If a server does not advertise block size constraints, it should
        # support these values. It can also return reads and block status info
        # aligned to minimum block size.
        # https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
        # section #block-size-constraints

        self.minimum_block_size = 1
        self.preferred_block_size = 4096
        self.maximum_block_size = 32 * MiB

        # Set to "qemu:dirty-bitmap:bitmap-name" if dirty is True, server
        # supports structued replies, and exports a dirty bitmap. Use this name
        # with extents() results to extract dirty extents.
        self.dirty_bitmap = None

        # Server capabilities discovered during handshake.
        self._structured_reply = False
        self._meta_context = {}

        self._counter = itertools.count()
        self._state = CONNECTING

        self._sock = self._connect(address)
        try:
            self._newstyle_handshake(dirty)
        except:  # noqa: E722
            self.close()
            raise

        log.debug("Ready for transmission")

    @property
    def has_base_allocation(self):
        return BASE_ALLOCATION in self._meta_context

    @property
    def has_allocation_depth(self):
        return QEMU_ALLOCATION_DEPTH in self._meta_context

    def read(self, offset, length):
        buf = bytearray(length)
        self.readinto(offset, buf)
        return buf

    def readinto(self, offset, buf):
        # If structured reply was negotiated, the server must send structured
        # reply to NBD_CMD_READ.
        cmd = Read(
            self._next_handle(), offset, buf,
            only_structured=self._structured_reply)
        self._send_command(cmd)
        self._recv_reply(cmd)
        return len(buf)

    def write(self, offset, data):
        cmd = Write(self._next_handle(), offset, len(data))
        self._send_command(cmd)
        self._send(data)
        self._recv_reply(cmd)

    def zero(self, offset, length, punch_hole=True):
        if self.transmission_flags & FLAG_SEND_WRITE_ZEROES == 0:
            raise UnsupportedRequest(
                "Server does not support CMD_WRITE_ZEROES")
        flags = 0 if punch_hole else CMD_FLAG_NO_HOLE
        cmd = WriteZeroes(self._next_handle(), offset, length, flags=flags)
        self._send_command(cmd)
        self._recv_reply(cmd)

    def flush(self):
        # TODO: is this the best way to handle this?
        if self.transmission_flags & FLAG_SEND_FLUSH == 0:
            return
        cmd = Flush(self._next_handle())
        self._send_command(cmd)
        self._recv_reply(cmd)

    def extents(self, offset, length):
        cmd = BlockStatus(self._next_handle(), offset, length)
        self._send_command(cmd)
        self._recv_reply(cmd)
        return cmd.reply

    def close(self):
        if self._state in (HANDSHAKE, TRANSMISSION):
            self._soft_disconnect()
        else:
            self._hard_disconnect()

    # Connecting to NBD server

    def _connect(self, address):
        """
        Connect to NBD server on address and return a connected socket, or
        raise socket.error.
        """
        if address.transport == "unix":
            return self._create_unix_connection(address)
        elif address.transport == "tcp":
            return self._create_tcp_connection(address)
        else:
            raise Error("Unsupported transport: {}".format(address))

    def _create_tcp_connection(self, address):
        """
        Enhanced version of socket.create_connection.

        Resolve DNS name to both AF_INET and AF_INET6 and will try to connect
        to all possible addresses.

        Set socket option TCP_NODELAY for improved latency.
        """
        sock = socket.create_connection(address)
        try:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except:  # noqa: E722
            sock.close()
            raise

        return sock

    def _create_unix_connection(self, address):
        """
        Like socket.create_connection() for unix socket.
        """
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(address)
        except:  # noqa: E722
            sock.close()
            raise

        return sock

    # NBD fixed newstyle handshake

    def _newstyle_handshake(self, dirty=False):
        assert self._state == CONNECTING
        self._state = HANDSHAKE

        # Initial handshake.
        nbd_magic, cliserv_magic, server_flags = self._recv_fmt("!QQH")

        if nbd_magic != NBDMAGIC:
            raise ProtocolError("Bad nbd magic {:x}, expecting {:x}"
                                .format(nbd_magic, NBDMAGIC))

        if cliserv_magic != IHAVEOPT:
            raise ProtocolError("Server does not support newstyle negotiation "
                                "magic={:x} expected={:x}"
                                .format(cliserv_magic, IHAVEOPT))

        log.debug("Received server flags: %x", server_flags)
        if not server_flags & FLAG_FIXED_NEWSTYLE:
            raise ProtocolError(
                "Server does not support fixed newstyle negotiation")

        self._send_client_flags(FLAG_C_FIXED_NEWSTYLE)

        # Options haggling.

        self._negotiate_structured_reply_option()

        if self._structured_reply:
            dirty_bitmap = self._query_dirty_bitmap() if dirty else None
            self._set_meta_context(dirty_bitmap)

        self._negotiate_go_option()

        self._state = TRANSMISSION

    def _send_client_flags(self, flags):
        log.debug("Sending client flags: %x", flags)
        b = struct.pack("!I", flags)
        self._send(b)

    # Negotiating options

    def _negotiate_structured_reply_option(self):
        """
        Ask the server to enable structured replies. This allows better error
        handling for CMD_READ, and enables extension that require
        structured replies such as CMD_BLOCK_STATUS.

        If negotiation was successful, the server MUST use structured reply to
        any response with a payload, and may used structured reply for other
        responses.

        If the server fails with REP_ERR_UNSUP, we disable structured
        replies and will not be able to report block status.
        """
        try:
            self._negotiate_option(OPT_STRUCTURED_REPLY)
        except OptionUnsupported as e:
            log.warning("Structured reply is not available: %s", e)
        else:
            log.debug("Structured reply enabled")
            self._structured_reply = True

    def _query_dirty_bitmap(self):
        """
        Query the server for dirty bitmap and return the context name if the
        server exports one, or None if server does not export any, or exports
        more than one.
        """
        opt = OPT_LIST_META_CONTEXT
        data = self._format_meta_context_data(QEMU_DIRTY_BITMAP)
        self._send_option(opt, data)

        bitmaps = list(self._iter_meta_context_replies(opt))

        if len(bitmaps) == 0:
            log.warning(
                "Server does not support %s meta context", QEMU_DIRTY_BITMAP)
            return None

        if len(bitmaps) > 1:
            log.warning("Cannot use multiple dirty bitmaps: %s", bitmaps)
            return None

        ctx_name, _ = bitmaps[0]
        log.debug("Server has dirty bitmap %s", ctx_name)
        return ctx_name

    def _set_meta_context(self, dirty_bitmap=None):
        """
        Register wanted meta context with the server.
        """
        opt = OPT_SET_META_CONTEXT

        # qemu:allocation-depth is required to detect holes in qcow2 images -
        # unallocated clusters exposing data from the backing chain.
        # Added in qemu 5.2.0.
        queries = [BASE_ALLOCATION, QEMU_ALLOCATION_DEPTH]
        if dirty_bitmap:
            queries.append(dirty_bitmap)

        data = self._format_meta_context_data(*queries)
        self._send_option(opt, data)

        # If the server supports OPT_SET_META_CONTEXT and all the contexts
        # in self._meta_context, we expect to get one reply for every context,
        # with the meta context id, and then REP_ACK.
        #
        # If the server does not support meta context, we may get
        # REP_ERR_UNSUP. If the server supports OPT_SET_META_CONTEXT
        # but not all required meta contexts, we may get info only about the
        # contexts supported by the server.
        #
        # Related sections in the spec:
        # - https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
        #   #metadata-querying
        # - https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
        #   #option-types (see OPT_SET_META_CONTEXT).

        for ctx_name, ctx_id in self._iter_meta_context_replies(opt):
            if ctx_name not in queries:
                raise ProtocolError(
                    "Unexpected context {}, expecting one of {}"
                    .format(ctx_name, queries))

            log.debug("Meta context %s is available id=%s", ctx_name, ctx_id)

            # Keep also reverse mapping to find name from id.
            self._meta_context[ctx_name] = ctx_id
            self._meta_context[ctx_id] = ctx_name

            if ctx_name == dirty_bitmap:
                self.dirty_bitmap = dirty_bitmap

        # Log if some meta context is not available. This may affect
        # performance or reduce funcionaliity, but is expected when using old
        # qemu-nbd (e.g. 4.2.0) or when using raw volume that does not provide
        # interesting allocation depth, and triggers a bug in qemu 6.2.0.
        for ctx_name in queries:
            if ctx_name not in self._meta_context:
                log.info("Meta context %s is not available", ctx_name)

    def _format_meta_context_data(self, *queries):
        """
        32 bits, length of export name.
        String, name of export for which we wish to list metadata contexts.
        32 bits, number of queries
        Zero or more queries, each being:
            32 bits, length of query
            String, query to select metadata contexts.
        """
        # Export name (length + name)
        name = self.export_name.encode("utf-8")
        data = bytearray()
        data += struct.pack("!I", len(name))
        data += name

        # Queries
        data += struct.pack("!I", len(queries))
        for query in queries:
            query = query.encode("utf-8")
            data += struct.pack("!I", len(query))
            data += query

        return data

    def _iter_meta_context_replies(self, opt):
        """
        Receive replies for NBD_LIST_META_CONTEXT and NBD_OPT_SET_META_CONTEXT
        commands. We assume that the replies and errors are identical.
        """
        while True:
            reply, length = self._recv_option_reply(opt)

            if reply in ERROR_REPLY:
                try:
                    self._handle_option_error(opt, reply, length)
                except OptionUnsupported as e:
                    log.warning("Meta context is not supported: %s", e)
                    return

            if reply == REP_ACK:
                if length != 0:
                    raise InvalidLength(reply, length, 0)
                break

            if reply != REP_META_CONTEXT:
                raise UnexpectedOptionReply(reply, opt, REP_META_CONTEXT)

            yield self._recv_meta_context_reply(length)

    def _recv_meta_context_reply(self, length):
        """
        Receive reply to OPT_SET_META_CONTEXT, and store the meta context
        id in self._meta_context dict.

        32 bits, NBD metadata context ID.
        String, name of the metadata context. This is not required to be a
            human-readable string, but it MUST be valid UTF-8 data.
        """
        if length < 4:
            raise InvalidLength(REP_META_CONTEXT, length, ">= 4")

        data = self._recv(length)
        ctx_id = struct.unpack("!I", data[:4])[0]
        ctx_name = data[4:].decode("utf-8")

        return ctx_name, ctx_id

    def _negotiate_go_option(self):
        # Here we can announce that we can honour server block size constraints
        # by adding INFO_BLOCK_SIZE information request. If we do this we
        # MUST abide by the block size constraints received. If we don't we are
        # allowed to send unaligned requests.
        # https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
        # section #block-size-constraints

        opt = OPT_GO
        data = self._format_go_option_data()
        self._send_option(opt, data)

        while True:
            reply, length = self._recv_option_reply(opt)

            if reply in ERROR_REPLY:
                self._handle_option_error(opt, reply, length)

            if reply == REP_ACK:
                if length != 0:
                    raise InvalidLength(reply, length, 0)
                if self.export_size is None or self.transmission_flags is None:
                    raise ProtocolError("Server did not send export size or "
                                        "transmission flags")
                break

            if reply != REP_INFO:
                raise UnexpectedOptionReply(reply, opt, REP_INFO)

            if length < 2:
                raise InvalidLength(REP_INFO, length, ">= 2")

            info = self._recv_fmt("!H")[0]
            length -= 2

            if info == INFO_EXPORT:
                self._recv_export_info(length)
            elif info == INFO_BLOCK_SIZE:
                self._recv_blocksize_info(length)
            else:
                data = self._recv(length)
                log.warning("Dropping unknown info reply=%r data=%r",
                            info, data)

    def _format_go_option_data(self, *requests):
        """
        Format export name and optional list of INFO_XXX requests.

        32 bits, length of name (unsigned); MUST be no larger than the option
            data length - 6
        String: name of the export
        16 bits, number of information requests
        16 bits x n - list of INFO information requests
        """
        # Export name (length + name)
        name = self.export_name.encode("utf-8")
        data = bytearray()
        data += struct.pack("!I", len(name))
        data += name

        # Information requests list (length + requests)
        data += struct.pack("!H", len(requests))
        if requests:
            data += struct.pack("!%dH" % len(requests), *requests)

        return data

    def _recv_export_info(self, length):
        if length != 10:
            raise InvalidLength(REP_INFO, length, 10)
        self.export_size, self.transmission_flags = self._recv_fmt("!QH")
        log.debug("Received export info size=%r flags=%r",
                  self.export_size, self.transmission_flags)

    def _recv_blocksize_info(self, length):
        if length != 12:
            raise InvalidLength(REP_INFO, length, 12)
        (self.minimum_block_size, self.preferred_block_size,
            self.maximum_block_size) = self._recv_fmt("!III")
        log.debug("Received block size info minimum=%r preferred=%r "
                  "maximum=%r",
                  self.minimum_block_size,
                  self.preferred_block_size,
                  self.maximum_block_size)

    # Negotiating options

    def _negotiate_option(self, opt, data=b""):
        self._send_option(opt, data)
        reply, length = self._recv_option_reply(opt)

        if reply in ERROR_REPLY:
            self._handle_option_error(opt, reply, length)

        # The spec is not clear about the possible reply for general options.
        # using qemu policy as in nbd_request_simple_option().

        if reply != REP_ACK:
            raise UnexpectedOptionReply(reply, opt, REP_ACK)

        if length != 0:
            raise InvalidLength(reply, length, 0)

    def _send_option(self, opt, data=b""):
        """
        Send an option with optional data to the server. The caller must call
        _recv_option_reply() to get a reply.
        """
        log.debug("Sending option: %r data: %r", opt, data)
        b = OPTION.pack(IHAVEOPT, opt, len(data))
        self._send(b)
        if data:
            self._send(data)

    def _recv_option_reply(self, expected_option):
        """
        Receive reply header from server, and return the reply and the length
        of the data that the caller need to read from the server to complete
        the option negotiation.

        S: 64 bits, 0x3e889045565a9 (magic number for replies)
        S: 32 bits, the option as sent by the client to which this is a reply
        S: 32 bits, reply type:
            - REP_ACK for successful completion, or
            - REP_ERR_UNSUP option not known by this server
        S: 32 bits, length of the reply; if zero, next field is not sent
        S: any data as required by the reply.
        """
        magic, option, reply, length = self._recv_fmt("!QIII")
        log.debug("Received reply magic=%x option=%s type=%s len=%s",
                  magic, option, reply, length)

        if magic != OPTION_REPLY_MAGIC:
            raise ProtocolError(
                "Unexpected reply magic {:x} for option {}, expecting {:x}"
                .format(magic, expected_option, OPTION_REPLY_MAGIC))

        if option != expected_option:
            raise ProtocolError("Unexpected reply option {}, expecting {}"
                                .format(option, expected_option))

        return reply, length

    def _handle_option_error(self, opt, reply, length):
        """
        Consume the optional data which is an error message suitable for
        displaying to the user, and raise an OptionError.
        """
        message = ""

        # If the server sent an error message, try to use it.
        if length:
            message = self._recv(length).decode("utf-8", errors="replace")

        # If we have no message, use the builtin message for this error.
        if message == "":
            message = ERROR_REPLY.get(reply, "Unknown error")

        if reply == REP_ERR_UNSUP:
            raise OptionUnsupported(opt, message)
        else:
            raise OptionError(opt, reply, message)

    # Terminating session

    def _soft_disconnect(self):
        """
        Perform soft disconnect.

        During handshake, we need to send a OPT_ABORT. The server
        may reply, but we are allowed to close the socket without
        reading the reply[1].

        During transmission, we need to send a CMD_DISC. The
        server does not reply[2].

        [1] https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
            #termination-of-the-session-during-option-haggling
        [2] https://github.com/NetworkBlockDevice/nbd/blob/master/doc/proto.md
            #terminating-the-transmission-phase
        """
        log.debug("Initiating a soft disconnect")
        try:
            if self._state == HANDSHAKE:
                self._send_option(OPT_ABORT)
            elif self._state == TRANSMISSION:
                cmd = Disc(self._next_handle())
                self._send_command(cmd)
            else:
                raise AssertionError(
                    "Cannot initiate soft disconnect at state {!r}"
                    .format(self._state))
        except socket.error as e:
            log.debug("Error initiating soft disconnect: %s", e)
        except Exception:
            log.exception("Error initiating soft disconnect")

        self._state = CLOSED
        self._close_socket()

    def _hard_disconnect(self):
        if self._state < CLOSED:
            log.debug("Initiating a hard disconnect")
            self._state = CLOSED
            self._close_socket()

    def _close_socket(self):
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except socket.error as e:
            log.debug("Error shutting down socket: %s", e)
        self._sock.close()

    # Commands

    def _next_handle(self):
        return next(self._counter)

    def _send_command(self, cmd):
        log.debug("Sending %s", cmd)
        self._send(cmd.to_bytes())

    def _recv_reply(self, cmd):
        """
        Receive either a simple reply or structured reply for cmd.
        """
        while True:
            magic = self._recv_fmt("!I")[0]

            if magic == SIMPLE_REPLY_MAGIC:
                if cmd.only_structured:
                    raise ProtocolError(
                        "Unexpected simple reply magic {:x}, expecting "
                        "structured reply magic {:x}"
                        .format(magic, STRUCTURED_REPLY_MAGIC))

                self._recv_simple_reply(cmd)
                return

            elif magic == STRUCTURED_REPLY_MAGIC:
                if not self._structured_reply:
                    raise ProtocolError(
                        "Unexpected structured reply magic {:x}, expecting "
                        "simple reply magic {:x}"
                        .format(magic, SIMPLE_REPLY_MAGIC))

                # We started to received structured reply chunks, so simple
                # reply is not allowed.
                cmd.only_structured = True

                if self._recv_reply_chunk(cmd):
                    break

            else:
                raise ProtocolError("Unexpected reply magic {:x}"
                                    .format(magic))

        if cmd.errors:
            # Some chunks failed. We don't have a good way to report
            # partial failures since content chunks may be fragmented, so
            # fail the entire request.
            raise RequestError("Errors receiving reply: {}".format(cmd.errors))

    def _recv_simple_reply(self, cmd):
        """
        Receive a simple reply (magic was already read).

        S: 32 bits, error (MAY be zero)
        S: 64 bits, handle
        S: (length bytes of data if the request is of type CMD_READ and
           error is zero)
        """
        error, handle = self._recv_fmt("!IQ")

        if error != 0:
            # We have no context in this case.
            raise ReplyError(error, "Simple reply failed")

        if handle != cmd.handle:
            raise UnexpectedHandle(handle, cmd.handle)

        if cmd.buf:
            self._recv_into(cmd.buf)

    def _recv_reply_chunk(self, cmd):
        """
        Receive a structured reply chunk (magic was already read). Return True
        if this was the last chunk.

        S: 16 bits, flags
        S: 16 bits, type
        S: 64 bits, handle
        S: 32 bits, length of payload (unsigned)
        S: length bytes of payload data (if length is nonzero)
        """
        flags, type, handle, length = self._recv_fmt("!HHQI")

        if handle != cmd.handle:
            raise UnexpectedHandle(handle, cmd.handle)

        if type == REPLY_TYPE_ERROR:
            self._handle_error_chunk(length, flags)

        if type == REPLY_TYPE_ERROR_OFFSET:
            self._handle_error_offset_chunk(length, cmd)
        elif type == REPLY_TYPE_NONE:
            self._handle_none_chunk(flags, length)
        elif type == REPLY_TYPE_OFFSET_DATA:
            self._handle_data_chunk(length, cmd)
        elif type == REPLY_TYPE_OFFSET_HOLE:
            self._handle_hole_chunk(length, cmd)
        elif type == REPLY_TYPE_BLOCK_STATUS:
            self._handle_block_status_chunk(length, cmd)
        else:
            raise ProtocolError(
                "Received unknown chunk type={} flags={} length={}"
                .format(type, flags, length))

        return flags & REPLY_FLAG_DONE

    def _handle_block_status_chunk(self, length, cmd):
        """
        Receive block status chunk and populate cmd's reply dict.

        The payload starts with 32 bits, metadata context ID
        and is followed by a list of one or more Extent descriptors.
        """
        ctx_id_size = 4
        extents_count, reminder = divmod(length, Extent.size)

        if extents_count == 0 or reminder != ctx_id_size:
            raise ProtocolError(
                "Received invalid payload length {}"
                .format(length))

        if extents_count > MAX_EXTENTS:
            raise ProtocolError(
                "Received too many extents {} > {}"
                .format(extents_count, MAX_EXTENTS))

        ctx_id = self._recv_fmt("!I")[0]

        ctx_name = self._meta_context.get(ctx_id)
        if ctx_name is None:
            raise ProtocolError(
                "Received unexpected metadata context id {}".format(ctx_id))

        if ctx_name == self.dirty_bitmap:
            context = Extent.DIRTY
        elif ctx_name == QEMU_ALLOCATION_DEPTH:
            context = Extent.DEPTH
        else:
            context = Extent.ALLOC

        extents = []
        for ext in self._recv_extents(length - ctx_id_size, context):
            if ext.length % self.minimum_block_size:
                raise ProtocolError(
                    "Invalid extent length {}: not an integer multiple "
                    "of minimum block size {}"
                    .format(ext.length, self.minimum_block_size))
            extents.append(ext)

        cmd.reply[ctx_name] = extents

    def _recv_extents(self, length, context):
        """
        Iterator receiving and unpacking extents descriptors form block status
        payload.

        Yield Extent object for every extent descriptor.
        """
        # We don't expect many extents in 4 GiB, so 1024 extents per call
        # should be more than enough.
        buf = bytearray(min(length, Extent.size * 1024))

        while length:
            # Shrink buffer for the last receive.
            if length < len(buf):
                buf = memoryview(buf)[:length]

            # Receive next buffer.
            self._recv_into(buf)
            length -= len(buf)

            # Unpack extents in buffer.
            view = memoryview(buf)
            while len(view):
                yield Extent.unpack(view[:Extent.size], context)
                view = view[Extent.size:]

    def _handle_none_chunk(self, flags, length):
        if not flags & REPLY_FLAG_DONE:
            raise ProtocolError(
                "Invalid none reply chunk without done flag type={} flags={}"
                .format(REPLY_TYPE_NONE, flags))
        if length != 0:
            raise InvalidLength(REPLY_TYPE_NONE, length, 0)

    def _handle_error_chunk(self, length, flags):
        """
        Handle general error (entire request failed).

        If this the last chunk raise ReplyError, failing this request.
        Otherwise raise ProtocolError failing entire connection.

        32 bits: error (MUST be nonzero)
        16 bits: message length (no more than header length - 6)
        message length bytes: optional string suitable for direct display to a
            human being
        """
        code, message = self._recv_error_chunk(length)

        if flags & REPLY_FLAG_DONE:
            raise ReplyError(code, message)
        else:
            raise ProtocolError(
                "Unrecoverable error chunk code={} message={!r}"
                .format(code, message))

    def _handle_error_offset_chunk(self, length, cmd):
        """
        Handle error at offset (partial error). This may not be the last chunk,
        so we collect the error and continue to read the next chunk.

        32 bits: error (MUST be nonzero)
        16 bits: message length (no more than header length - 14)
        message length bytes: optional string suitable for direct display to a
            human being
        64 bits: offset (unsigned)
        """
        code, message = self._recv_error_chunk(length - 8)
        offset = self._recv_fmt("!Q")[0]
        cmd.errors.append((offset, ReplyError(code, message)))

    def _handle_data_chunk(self, length, cmd):
        """
        Receive data chunk payload into cmd's buf.

        64 bits: offset (unsigned)
        length - 8 bytes: data
        """
        # TODO: Validate that chunk offset and size are within requested range.
        chunk_offset = self._recv_fmt("!Q")[0]
        chunk_size = length - 8

        log.debug("Receive data chunk offset=%s size=%s",
                  chunk_offset, chunk_size)

        buf_offset = chunk_offset - cmd.offset
        with memoryview(cmd.buf)[buf_offset:buf_offset + chunk_size] as view:
            self._recv_into(view)

    def _handle_hole_chunk(self, length, cmd):
        """
        Handle hole chunk, zeroing byte range in cmd's buf.

        64 bits: offset (unsigned)
        32 bits: hole size (unsigned, MUST be nonzero)
        """
        if length != 12:
            raise InvalidLength(REPLY_TYPE_OFFSET_HOLE, length, 12)

        chunk_offset, chunk_size = self._recv_fmt("!QI")
        if chunk_size == 0:
            raise ProtocolError("Invalid hole chunk with zero size")

        log.debug("Receive hole chunk offset=%s size=%s",
                  chunk_offset, chunk_size)

        buf_offset = chunk_offset - cmd.offset
        cmd.buf[buf_offset:buf_offset + chunk_size] = b"\0" * chunk_size

    def _recv_error_chunk(self, length):
        code, msg_len = self._recv_fmt("!IH")

        if msg_len != length - 6:
            raise ProtocolError(
                "Invalid structure reply error message length {}, expected={}"
                .format(msg_len, length - 6))

        # The protocol does not specify the encoding.
        message = self._recv(msg_len).decode("utf-8", errors="replace")
        return code, message

    # Structured I/O

    def _recv_fmt(self, fmt):
        s = struct.Struct(fmt)
        data = self._recv(s.size)
        return s.unpack(data)

    # Plain I/O

    def _send(self, data):
        self._sock.sendall(data)

    def _recv(self, length):
        buf = bytearray(length)
        self._recv_into(buf)
        return buf

    def _recv_into(self, buf):
        with memoryview(buf) as view:
            length = len(view)
            pos = 0
            while pos < length:
                n = self._sock.recv_into(view[pos:])
                if not n:
                    raise ProtocolError(
                        "Server closed the connection, read {} bytes, "
                        "expected {} bytes"
                        .format(pos, length))
                pos += n

    # Conetext manager

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            # Don't hide excpetions in user code.
            if t is None:
                raise
            log.exeption("Error closing")


class Command:
    """
    Abstract NBD command.

    All commands share the same wire format:

        32 bits, 0x25609513, magic (REQUEST_MAGIC)
        16 bits, command flags
        16 bits, type
        64 bits, handle
        64 bits, offset (unsigned)
        32 bits, length (unsigned)
    """

    wire_format = struct.Struct("!IHHQQI")

    # Attributes defined by sub classes.
    name = None
    type = None
    buf = None

    def __init__(self, handle, offset=0, length=0, flags=0):
        self.handle = handle
        self.offset = offset
        self.length = length
        self.flags = flags
        self.only_structured = False
        # NBD_REPLY_TYPE_ERROR_OFFSET chunks received when handling structued
        # reply. Can happen only in Read, Write, and BlockStatus.
        self.errors = []

    def to_bytes(self):
        return self.wire_format.pack(
            REQUEST_MAGIC,
            self.flags,
            self.type,
            self.handle,
            self.offset,
            self.length)

    def __str__(self):
        return "{} handle={} offset={} length={} flags={}".format(
            self.name, self.handle, self.offset, self.length, self.flags)


class Read(Command):
    type = 0
    name = "NBD_CMD_READ"

    def __init__(self, handle, offset, buf, only_structured=False):
        super().__init__(handle, offset, len(buf))
        # Buffer for storing the payload from the server.
        self.buf = buf
        self.only_structured = only_structured


class Write(Command):
    type = 1
    name = "NBD_CMD_WRITE"


class Disc(Command):
    type = 2
    name = "NBD_CMD_DISC"

    def __init__(self, handle):
        super().__init__(handle, 0, 0)


class Flush(Command):
    type = 3
    name = "NBD_CMD_FLUSH"

    def __init__(self, handle):
        super().__init__(handle, 0, 0)


class WriteZeroes(Command):
    type = 6
    name = "NBD_CMD_WRITE_ZEROES"


class BlockStatus(Command):
    type = 7
    name = "NBD_CMD_BLOCK_STATUS"

    def __init__(self, handle, offset, length):
        super().__init__(handle, offset, length)
        self.only_structured = True
        # Mapping of meta context name to list of Extent objects.
        self.reply = {}


class Extent:
    """
    A mutable extent of data or zeroes.

    The length field is mutable to allow merging consecutive extents received
    from the server, and splitting large extents when trimming extents that
    exceed the requested range.

    The flags field is read-only since we don't have a use case for changing
    it.

    Since this class is mutable, it must not implement __hash__.
    """

    # Contexts ids copied from qemu-nbd.
    ALLOC = 0  # base:allocation
    DEPTH = 1  # qemu:allocation-depth
    DIRTY = 2  # qemu:dirty-bitmap:name

    __slots__ = ("length", "_flags")

    # 32 bits, length of the extent to which the status below applies
    #     (unsigned, MUST be nonzero)
    # 32 bits, status flags
    wire_format = struct.Struct("!II")

    size = wire_format.size

    def __init__(self, length, flags):
        self.length = length
        self._flags = flags

    @classmethod
    def unpack(cls, data, context=ALLOC):
        """
        Create extent instance from extent data.

        Based on context, we map NBD flags bits to private bits to allow
        merging different types of extents.
        """
        length, flags = cls.wire_format.unpack(data)
        if length == 0:
            raise ProtocolError(
                "Invalid extent length=0 flags={}".format(flags))

        log.debug("Extent length=%s flags=%s context=%s",
                  length, flags, context)

        if context == cls.ALLOC:
            # The remainder of the flags field is reserved. Servers SHOULD set
            # it to all-zero; clients MUST ignore unknown flags.
            flags = flags & (STATE_HOLE | STATE_ZERO)
        elif context == cls.DIRTY:
            flags = EXTENT_DIRTY if flags & STATE_DIRTY else 0
        elif context == cls.DEPTH:
            flags = EXTENT_BACKING if flags == 0 else 0

        return cls(length, flags)

    @classmethod
    def pack(cls, length, flags):
        return cls.wire_format.pack(length, flags)

    @property
    def flags(self):
        return self._flags

    @property
    def zero(self):
        """
        For base:allocation extent, True if extents will read as zeros.
        """
        return bool(self._flags & STATE_ZERO)

    @property
    def hole(self):
        """
        Return True if this is a non existing extent in qcow2 image. If the
        image has a backing file the content of the extent are read from the
        backing file.

        Return False if the extent exist in the one of the layers, or if the
        image is a raw image, even if this is unallocated area in a raw
        image.

        By definition, a hole has also the STATE_HOLE | STATE_ZERO
        flags. But the oposite is not true.

        For more info see
        https://lists.nongnu.org/archive/html/qemu-block/2021-06/msg00756.html
        """
        return bool(self._flags & EXTENT_BACKING)

    @property
    def dirty(self):
        """
        For qemu:dirty-bitmap extent, True if extents was modified and is
        included in this incremental backup.
        """
        return bool(self._flags & EXTENT_DIRTY)

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.length == other.length and
                self.flags == other.flags)

    def __repr__(self):
        return "Extent(length={}, flags={})".format(self.length, self._flags)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import os

from . import errors
from . import http
from . import ssl
from . import stats
from . import uhttp
from . import util

from .handlers import (
    checksum,
    extents,
    images,
    info,
    profile,
    tickets,
)

DEFAULT_SOCKET_MODE = 0o660

log = logging.getLogger("services")


class Service:

    name = None

    def start(self):
        log.debug("Starting %s", self.name)
        util.start_thread(self._run, name=self.name)

    def stop(self):
        log.debug("Stopping %s", self.name)
        self._server.shutdown()

    @property
    def port(self):
        return self._server.server_port

    @property
    def address(self):
        return self._server.server_address

    def _run(self):
        log.debug("%s started", self.name)
        self._server.serve_forever(
            poll_interval=self._config.daemon.poll_interval)
        log.debug("%s terminated normally", self.name)


class RemoteService(Service):
    """
    Service used to access images data from remote host.

    Access to this service requires a valid ticket that can be installed using
    the local control service.
    """

    name = "remote.service"

    def __init__(self, config, auth):
        self._config = config
        port = config.remote.port
        if not 0 <= port < 0xFFFF:
            raise errors.InvalidConfig("remote.port", port)
        log.debug("Creating %s on port %d", self.name, port)
        self._server = http.Server((config.remote.host, port), http.Connection)
        # TODO: Make clock configurable, disabled by default.
        self._server.clock_class = stats.Clock
        if port == 0:
            config.remote.port = self.port
        if config.tls.enable:
            self._secure_server()
        self._server.app = http.Router([
            (r"/images/(.*)/extents", extents.Handler(config, auth)),
            (r"/images/(.*)/checksum/algorithms",
                checksum.Algorithms(config, auth)),
            (r"/images/(.*)/checksum", checksum.Checksum(config, auth)),
            (r"/images/(.*)", images.Handler(config, auth)),
            (r"/info/", info.Handler(config, auth)),
        ])
        log.info("%s listening on %r", self.name, self.address)

    def _secure_server(self):
        if "" in (self._config.tls.cert_file, self._config.tls.key_file):
            raise errors.TlsConfigurationError(self._config.tls)

        log.debug("Securing server (cafile=%s, certfile=%s, keyfile=%s)",
                  self._config.tls.ca_file,
                  self._config.tls.cert_file,
                  self._config.tls.key_file)
        context = ssl.server_context(
            self._config.tls.cert_file,
            self._config.tls.key_file,
            cafile=self._config.tls.ca_file,
            enable_tls1_1=self._config.tls.enable_tls1_1)
        self._server.socket = context.wrap_socket(
            self._server.socket, server_side=True)


class LocalService(Service):
    """
    Service used to access images locally.

    Access to this service requires a valid ticket that can be installed using
    the control service.
    """

    name = "local.service"

    def __init__(self, config, auth):
        self._config = config
        log.debug("Creating %s on socket %r", self.name, config.local.socket)
        self._server = uhttp.Server(config.local.socket, uhttp.Connection)
        # TODO: Make clock configurable, disabled by default.
        self._server.clock_class = stats.Clock
        if config.local.socket == "":
            config.local.socket = self.address
        self._server.app = http.Router([
            (r"/images/(.*)/extents", extents.Handler(config, auth)),
            (r"/images/(.*)/checksum/algorithms",
                checksum.Algorithms(config, auth)),
            (r"/images/(.*)/checksum", checksum.Checksum(config, auth)),
            (r"/images/(.*)", images.Handler(config, auth)),
        ])
        log.info("%s listening on %r", self.name, self.address)


class ControlService(Service):
    """
    Service used to control imageio daemon on a host.

    The service is using unix socket owned by a program managing the host. Only
    this program can access the socket.
    """

    name = "control.service"

    def __init__(self, config, auth):
        self._config = config
        transport = self._config.control.transport.lower()
        if transport == "tcp":
            port = config.control.port
            log.debug("Creating %s on port %r", self.name, port)
            if not 0 <= port < 0xFFFF:
                raise errors.InvalidConfig("control.port", port)
            self._server = http.Server(
                ("localhost", port),
                http.Connection,
                prefer_ipv4=config.control.prefer_ipv4)
            if port == 0:
                config.control.port = self.port
        elif transport == "unix":
            socket = config.control.socket
            log.debug("Creating %s on socket %r", self.name, socket)
            self._server = uhttp.Server(socket, uhttp.Connection)
            if socket == "":
                config.control.socket = self.address
            os.chmod(config.control.socket, DEFAULT_SOCKET_MODE)
        else:
            raise errors.InvalidConfig("control.transport", transport)

        # TODO: Make clock configurable, disabled by default.
        self._server.clock_class = stats.Clock

        self._server.app = http.Router([
            (r"/tickets/(.*)", tickets.Handler(config, auth)),
            (r"/profile/", profile.Handler(config, auth)),
        ])
        log.info("%s listening on %r", self.name, self.address)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import grp
import json
import os
import pwd
import subprocess

import pytest

from contextlib import contextmanager

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server
from ovirt_imageio._internal import sockutil

from . import http

DAEMON_CONFIG = """\
[daemon]
poll_interval = 0.1
buffer_size = 131072
run_dir = {run_dir}
drop_privileges = {drop_priv}
user_name = {user_name}
group_name = {group_name}

[tls]
key_file = test/pki/system/key.pem
cert_file = test/pki/system/cert.pem
ca_file = test/pki/system/ca.pem

[remote]
port = 0

[local]
socket =

[control]
transport = unix
socket = {run_dir}/sock

[logger_root]
level=DEBUG

[handler_logfile]
args=('{log_dir}/daemon.log',)
"""

requires_root = pytest.mark.skipif(os.geteuid() != 0, reason="Requires root")
requires_unprivileged = pytest.mark.skipif(
    os.geteuid() == 0, reason="Requires unprivileged user")


@pytest.fixture
def tmp_dirs(tmpdir):
    tmpdir.mkdir("run")
    tmpdir.mkdir("log")
    tmpdir.mkdir("conf").mkdir("conf.d")


def test_load_config(monkeypatch, tmpdir):
    # Here we test full scenario, when config is loaded from multiple sources
    # and test, that specified options were overwritten as expected.
    # Install config overwrites the default settings, then vendor config should
    # be loaded and overwrite log handler level and finally admin config should
    # be loaded and overwrite setup of control service previously defined by
    # install config.

    install_config = """
[control]
transport = unix
socket = test/daemon.sock

[handler_logfile]
class = logging.StreamHandler
args = ()
kwargs = {}
"""

    vendor_config = """
[handler_logfile]
level = ERROR
"""

    admin_config = """
[control]
transport = tcp
port = 10000
"""

    etc_dir = tmpdir.mkdir("etc")
    etc_conf_d = etc_dir.mkdir("conf.d")
    install_cfg = etc_conf_d.join("50-install.conf")
    install_cfg.write(install_config)
    admin_cfg = etc_conf_d.join("90-admin.conf")
    admin_cfg.write(admin_config)

    vendor_dir = tmpdir.mkdir("vendor")
    vendor_cfg = vendor_dir.mkdir("conf.d").join("60-vendor.conf")
    vendor_cfg.write(vendor_config)

    monkeypatch.setattr(server, "VENDOR_CONF_DIR", str(vendor_dir))
    cfg = server.load_config(str(etc_dir))

    assert cfg.control.transport == "tcp"
    assert cfg.control.port == 10000
    assert cfg.handler_logfile.level == "ERROR"
    assert cfg.handler_logfile.keyword__class == "logging.StreamHandler"


def test_load_config_missing():
    # Fail fast if conf_dir does not exist.
    with pytest.raises(ValueError):
        server.load_config("/no/such/dir")


def test_load_config_no_conf_d(tmpdir):
    # Fail fast if conf_dir does not contain a cond.d sub directory.
    conf_dir = tmpdir.mkdir("user")
    with pytest.raises(ValueError):
        server.load_config(str(conf_dir))


def test_load_config_no_conf_files(tmpdir):
    # Fail fast if conf_dir/cond.d/ does not contain any configurtion file.
    conf_dir = tmpdir.mkdir("user")
    conf_d = conf_dir.mkdir("cond.d")

    # Create a disabled configuration file, it will be ignored.
    disabled_cfg = conf_d.join("99-user.conf.disabled")
    disabled_cfg.write("[control]\ntransport = tcp\n")

    with pytest.raises(ValueError):
        server.load_config(str(conf_dir))


def test_show_config():
    cfg = config.load(["test/conf.d/daemon.conf"])
    out = subprocess.check_output(
        ["./ovirt-imageio", "--conf-dir", "./test", "--show-config"])
    assert json.loads(out) == config.to_dict(cfg)


@requires_unprivileged
def test_unprivileged_user(tmpdir, tmp_dirs):
    with started_imageio(tmpdir) as p:
        uid = os.getuid()
        gid = os.getuid()
        status = process_status(p.pid)
        assert status["uids"] == {uid}
        assert status["gids"] == {gid}
        assert status["groups"] == user_groups(uid)

        # Since we run as unprivileged users the daemon run as the same
        # user and group and no ownership changes are made.
        expected_user = pwd.getpwuid(os.getuid()).pw_name
        expected_group = grp.getgrgid(os.getuid()).gr_name
        assert_ownership(tmpdir, expected_user, expected_group)


@requires_root
def test_drop_privileges_disable(tmpdir, tmp_dirs):
    with started_imageio(tmpdir, drop_privileges="false") as p:
        # Run under root and privileges shouldn't be dropped and ownership of
        # files shouldn't be changed.
        status = process_status(p.pid)
        assert status["uids"] == {0}
        assert status["gids"] == {0}

        # Asserting groups directly to root groups fails on Jenkins as the test
        # probably run there under a mock user which inherits some groups.
        # Therefore we assert that the groups are same as groups of the parent
        # process.
        parent_process_status = process_status(os.getpid())
        assert status["groups"] == parent_process_status["groups"]
        assert_ownership(tmpdir, "root", "root")


@requires_root
def test_drop_privileges(tmpdir, tmp_dirs):
    with started_imageio(tmpdir) as p:
        # Run under root but privileges should be dropped, daemon should run
        # under nobody user and relevant files should be owned by this user.
        status = process_status(p.pid)
        pwnam = pwd.getpwnam("nobody")
        assert status["uids"] == {pwnam.pw_uid}
        assert status["gids"] == {pwnam.pw_gid}
        assert status["groups"] == user_groups("nobody")
        assert_ownership(tmpdir, "nobody", "nobody")


def prepare_config(tmpdir, drop_privileges="true"):
    daemon_conf = DAEMON_CONFIG.format(
        run_dir=os.path.join(tmpdir, "run"),
        log_dir=os.path.join(tmpdir, "log"),
        drop_priv=drop_privileges,
        user_name="nobody",
        group_name="nobody",
    )
    tmpdir.join("conf", "conf.d", "daemon.conf").write(daemon_conf)


@contextmanager
def started_imageio(tmpdir, drop_privileges="true"):
    prepare_config(tmpdir, drop_privileges=drop_privileges)

    conf_dir = tmpdir.join("conf")

    cmd = ["./ovirt-imageio", "--conf-dir", str(conf_dir)]
    proc = subprocess.Popen(cmd)
    try:
        socket = sockutil.UnixAddress(str(tmpdir.join("run", "sock")))
        if not sockutil.wait_for_socket(socket, 10):
            raise RuntimeError("Timeout waiting for {}".format(socket))

        # Wait until server is listening - at this point it already dropped
        # privileges.
        if drop_privileges:
            cfg = config.load(str(conf_dir.join("conf.d", "daemon.conf")))
            with http.ControlClient(cfg) as c:
                r = c.get("/tickets/no-such-ticket")
                r.read()
                assert r.status == 404

        yield proc
    finally:
        proc.terminate()
        proc.wait()


def user_groups(user):
    """
    Return group ids for username (str) or user id (int)
    """
    gids = subprocess.check_output(["id", "--groups", "-z", str(user)]).decode(
        "utf-8").strip("\0").split("\0")
    return {int(gid) for gid in gids}


def parse_ids(line):
    _, values = line.split(":", 1)
    return set(int(x) for x in values.strip().split())


def process_status(pid):
    """
    Returns uid, gid and groups of the process.
    Example for the the process with following status:

        $ cat /proc/2769/status | egrep '^(Uid:|Gid:|Groups:)'
        Uid:    993     993     993     993
        Gid:    990     990     990     990
        Groups: 990

    the output is:

        > process_status(2769)
        {'uids': {993}, 'gids': {990}, 'groups': {990}}
    """
    status_path = os.path.join("/proc", str(pid), "status")
    with open(status_path, "r") as f:
        status_lines = f.readlines()

    for line in status_lines:
        if line.startswith("Uid:"):
            uids = parse_ids(line)
        if line.startswith("Gid:"):
            gids = parse_ids(line)
        if line.startswith("Groups:"):
            groups = parse_ids(line)

    return {"uids": uids, "gids": gids, "groups": groups}


def assert_path_owner(path, expected_user, expected_group):
    path_stat = os.stat(path)
    uid = path_stat.st_uid
    gid = path_stat.st_gid
    user = pwd.getpwuid(uid).pw_name
    group = grp.getgrgid(gid).gr_name

    assert expected_user == user
    assert expected_group == group


def assert_ownership(tmpdir, user, group):
    run_dir = os.path.join(tmpdir, "run")
    assert_path_owner(run_dir, user, group)
    assert_path_owner(str(tmpdir.join("run", "sock")), user, group)
    assert_path_owner(str(tmpdir.join("log", "daemon.log")), user, group)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
nbdutil - Network Block Device utility functions.

Geting extents
==============

NBD spec is very liberal about returning extents. A client must handle these
cases when handling reply for block status command.

Single extent
-------------

The server is allowed to return extent for any request, regardless of the
actual extents on storage. In this case the client must send more block status
commands to get the rest of the extent for the requested range.

Storage:  [-------------|-------------|---------------|

Request:  [----------------------------------]
Reply:    [-------------]


Request:                [--------------------]
Reply:                  [-------------]

Request:                              [------]
Reply:                                [------]

Result:   [-------------|-------------|------]


Short reply
-----------

The server can return some extents, not covering the entire requested ragne.
This is mostly like single extent case.

Storage:  [-------------|-------------|---------------|

Request:  [----------------------------------]
Reply:    [-------------]-------------]

Request:                              [------]
Reply:                                [------]

Result:   [-------------|-------------|------]


Long reply
----------

When the server return multiple extents (N), the N-1 extent must be within the
requested ragne, but the last extent may exceed it. In this case the client
need to clip the last extent to the requested range.

Storage:  [-------------|-------------|---------------|

Request:  [----------------------------------]
Reply:    [-------------]-------------|---------------]

Result:   [-------------|-------------|------]


Consecutive extents of same type
--------------------------------

The server may return multiple extents of same type for the same extent on
storage. The client need to merge the consecutive extents.

Storage:  [xxxxxxxxxxxxx|0000000000000|xxxxxxxxxxxxxxx|

Request:  [----------------------------------]
Reply:    [xxxxxx|xxxxxx|000000|000000|xxxxxx]

Result:   [xxxxxxxxxxxxx|0000000000000|xxxxxx]

"""

import logging
import queue
import sys

from collections import namedtuple

from . import nbd
from . import util
from .units import MiB, GiB

# NBD spec allows zeroing up to 2**32 - 1 bytes, buf some nbd servers like
# qemu-nbd limit seems to be 2**31 - 512. Large zeros can delay more important
# I/O so we like to zero is smaller steps.
MAX_ZERO = GiB

log = logging.getLogger("nbdutil")


def extents(client, offset=0, length=None, dirty=False):
    """
    Iterate over all extents for requested range.

    Requested range must not excceed image size, but is not limited by NBD
    maximum length, since we send multiple block status commands to the server.

    Consecutive extents of same type are merged automatically.

    Return iterator of nbd.Extent objects.
    """
    if length is None:
        end = client.export_size
    else:
        end = offset + length

    # NBD limit extents request to 4 GiB - 1. We use smaller step to limit the
    # number of extents kept in memory when accessing very fragmented images.
    max_step = 2 * GiB

    # Keep the current extent, until we find a new extent with different flags.
    cur = None

    while offset < end:
        # Get the next extent reply since the last returned extent. This
        # handles the cases of single extent, short reply, and last extent
        # exceeding requested range.
        step = min(end - offset, max_step)
        res = client.extents(offset, step)

        if dirty:
            extents = merged(
                res[nbd.BASE_ALLOCATION], res[client.dirty_bitmap])
        elif nbd.QEMU_ALLOCATION_DEPTH in res:
            extents = merged(
                res[nbd.BASE_ALLOCATION], res[nbd.QEMU_ALLOCATION_DEPTH])
        else:
            extents = res[nbd.BASE_ALLOCATION]

        for ext in extents:
            # Handle the case of last extent of the last block status command
            # exceeding requested range.
            if offset + ext.length > end:
                ext.length = end - offset

            offset += ext.length

            # Handle the case of consecutive extents with same flags.
            if cur is None:
                cur = ext
            elif cur.flags == ext.flags:
                cur.length += ext.length
            else:
                yield cur
                cur = ext

            # The spec does not allow the server to send more extent. Ensure
            # that we don't report wrong data if the server does not comply.
            if offset == end:
                break

    yield cur


def merged(extents_a, extents_b):
    """
    Merge lists of extents with distinct flags bits, yielding merged extents
    with flags from both lists. Yielding ends when the first iterator is
    consumed.

    Yields nbd.Extent() including all bits from both extents.
    """
    iter_a = iter(extents_a)
    iter_b = iter(extents_b)

    a = None
    b = None

    while True:
        try:
            if a is None:
                a = next(iter_a)
            if b is None:
                b = next(iter_b)
        except StopIteration:
            return

        if a.length == b.length:
            # The easy case, merge and clear both.
            yield nbd.Extent(a.length, a.flags | b.flags)
            a = None
            b = None
        elif a.length > b.length:
            # Yield the overlapping area and keep rest of a.
            yield nbd.Extent(b.length, a.flags | b.flags)
            a = nbd.Extent(a.length - b.length, a.flags)
            b = None
        else:  # a.length < b.length
            # Yield the overlapping area and keep rest of b.
            yield nbd.Extent(a.length, a.flags | b.flags)
            b = nbd.Extent(b.length - a.length, b.flags)
            a = None


def copy(src_client, dst_client, block_size=4 * MiB, queue_depth=4,
         progress=None):
    """
    Copy export from src_client to dst_client.

    Both exports must have identical size, but can have different format.
    """

    # Consider both requested block size and clients limits.
    buf_size = min(
        block_size,
        min(src_client.maximum_block_size, dst_client.maximum_block_size))

    # Leave extra room for None buffer signaling that the writer failed.
    buffers = queue.Queue(queue_depth + 1)

    # Allocate buffers for write requests.
    for _ in range(queue_depth):
        buffers.put(bytearray(buf_size))

    # Keep room for queue_depth write requests (have buffer) and
    # queue_depth zero requests (have no buffer).
    requests = queue.Queue(queue_depth * 2)

    error = [None]

    log.debug("starting writer thread")
    writer = util.start_thread(
        _write,
        args=(dst_client, requests, buffers, error, progress),
        name="writer")

    _read(src_client, requests, buffers, error)

    log.debug("waiting for writer thread")
    writer.join()

    if error[0]:
        raise error[0][1]


# Request ops.
WRITE = "write"
ZERO = "zero"
FLUSH = "flush"


class Request(namedtuple("Request", "op,offset,length,buf")):
    __slots__ = ()

    def __new__(cls, op, offset=0, length=0, buf=None):
        return tuple.__new__(cls, (op, offset, length, buf))


def _read(client, requests, buffers, error):
    log.debug("reader started")

    offset = 0
    for ext in extents(client):
        todo = ext.length
        if ext.zero:
            while todo:
                if error[0]:
                    log.debug("reader stopped")
                    return

                step = min(todo, MAX_ZERO)
                requests.put(Request(ZERO, offset, step))
                offset += step
                todo -= step
        else:
            while todo:
                buf = buffers.get()
                if error[0]:
                    log.debug("reader stopped")
                    return

                step = min(todo, len(buf))
                view = memoryview(buf)[:step]
                client.readinto(offset, view)
                requests.put(Request(WRITE, offset, step, buf))
                offset += step
                todo -= step

    requests.put(Request(FLUSH))
    requests.put(None)
    log.debug("reader finished")


def _write(client, requests, buffers, error, progress=None):
    try:
        log.debug("writer started")

        while True:
            req = requests.get()
            if req is None:
                log.debug("writer stopped")
                break

            if req.op is ZERO:
                client.zero(req.offset, req.length)
            elif req.op is WRITE:
                view = memoryview(req.buf)[:req.length]
                client.write(req.offset, view)
                buffers.put(req.buf)
            elif req.op is FLUSH:
                client.flush()
            else:
                raise RuntimeError("Unknown request: {}".format(req))

            if progress:
                progress.update(req.length)

        log.debug("writer finished")
    except Exception:
        log.debug("writer failed")
        error[0] = sys.exc_info()
        buffers.put(None)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import time

import pytest

from ovirt_imageio._internal import util


def test_start_thread_args():
    result = []

    def f(*a):
        result.extend(a)

    util.start_thread(f, args=(1, 2)).join()
    assert result == [1, 2]


def test_start_thread_kwargs():
    result = {}

    def f(k=None):
        result["k"] = k

    util.start_thread(f, kwargs={"k": "v"}).join()
    assert result == {"k": "v"}


def test_start_thread_name():
    t = util.start_thread(lambda: None, name="foo")
    t.join()
    assert t.name == "foo"


def test_start_thread_daemon():
    t = util.start_thread(lambda: None)
    t.join()
    assert t.daemon


def test_start_thread_non_daemon():
    t = util.start_thread(lambda: None, daemon=False)
    t.join()
    assert not t.daemon


def test_monotonic_time():
    t1 = util.monotonic_time()
    time.sleep(0.01)
    t2 = util.monotonic_time()
    assert t1 <= t2


@pytest.mark.parametrize("size,rounded", [
    (0, 0),
    (1, 512),
    (512, 512),
    (512 + 1, 512 * 2),
])
def test_round_up(size, rounded):
    assert util.round_up(size, 512) == rounded


@pytest.mark.parametrize("size,rounded", [
    (0, 0),
    (1, 0),
    (512, 512),
    (512 + 1, 512),
])
def test_round_down(size, rounded):
    assert util.round_down(size, 512) == rounded


@pytest.mark.parametrize("value,expected", [
    ("value", "value"),
    ("value", "value"),
    (b"value", "value"),
    ("\u05d0", "\u05d0"),
    (b"\xd7\x90", "\u05d0"),
    ("\u0000", "\u0000"),
    (b"\0", "\u0000"),
])
def test_ensure_text(value, expected):
    result = util.ensure_text(value)
    assert isinstance(result, str)
    assert result == expected


def test_ensure_text_unexpected_type():
    with pytest.raises(TypeError):
        util.ensure_text(1)


def test_unbuffered_stream_more():
    chunks = [b"1" * 256,
              b"2" * 256,
              b"3" * 42,
              b"4" * 256]
    s = util.UnbufferedStream(chunks)
    # Chunk 1
    b = s.read(512)
    assert b == chunks[0]
    # Chunk 2
    b = s.read(512)
    assert b == chunks[1]
    # Chunk 3
    b = s.read(512)
    assert b == chunks[2]
    # Chunk 4
    b = s.read(512)
    assert b == chunks[3]
    # Empty
    b = s.read(512)
    assert b == b''
    b = s.read(512)
    assert b == b''


def test_unbuffered_stream_less():
    chunks = [b"1" * 256,
              b"2" * 256,
              b"3" * 42,
              b"4" * 256]
    s = util.UnbufferedStream(chunks)
    # Chunk 1
    b = s.read(128)
    assert b == chunks[0][:128]
    b = s.read(128)
    assert b == chunks[0][128:]
    # Chunk 2
    b = s.read(128)
    assert b == chunks[1][:128]
    b = s.read(128)
    assert b == chunks[1][128:]
    # Chunk 3
    b = s.read(128)
    assert b == chunks[2]
    # Chunk 4
    b = s.read(128)
    assert b == chunks[3][:128]
    b = s.read(128)
    assert b == chunks[3][128:]
    # Empty
    b = s.read(128)
    assert b == b''
    b = s.read(128)
    assert b == b''


@pytest.mark.parametrize("n,s", [
    (0, "0 bytes"),
    (0.0, "0 bytes"),
    (1023, "1023 bytes"),
    (1024, "1.00 KiB"),
    (1024 * 1023, "1023.00 KiB"),
    (1024 * 1024, "1.00 MiB"),
    (1024**2 * 1023, "1023.00 MiB"),
    (1024**2 * 1024, "1.00 GiB"),
    (1024**3 * 1023, "1023.00 GiB"),
    (1024**3 * 1024, "1.00 TiB"),
    (1024**4 * 1023, "1023.00 TiB"),
    (1024**4 * 1024, "1.00 PiB"),
])
def test_humansize(n, s):
    assert util.humansize(n) == s



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import getpass
import os
import uuid

import pytest

from ovirt_imageio._internal.units import KiB, MiB, GiB, TiB
from ovirt_imageio._internal import version
from ovirt_imageio.client import _options


@pytest.mark.parametrize("value,size", [
    (42, "42"),
    (256 * KiB, "256k"),
    (32 * MiB, "32m"),
    (10 * GiB, "10g"),
    (1 * TiB, "1t"),
])
def test_sizevalue(value, size):
    sv = _options.SizeValue(value)
    assert sv == value
    assert str(sv) == size


@pytest.mark.parametrize("size,value", [
    ("0", 0),
    ("42", 42),
    ("256k", 256 * KiB),
    ("256K", 256 * KiB),
    ("32m", 32 * MiB),
    ("32M", 32 * MiB),
    ("10g", 10 * GiB),
    ("10G", 10 * GiB),
    ("1t", 1 * TiB),
    ("1T", 1 * TiB),
])
def test_size(size, value):
    validate = _options.Size()
    assert validate(size) == value


def test_size_invalid():
    validate = _options.Size()
    with pytest.raises(ValueError) as e:
        validate("not a number")
    assert repr("not a number") in str(e.value)


def test_size_negative():
    validate = _options.Size()
    with pytest.raises(ValueError) as e:
        validate("-1")
    assert repr("-1") in str(e.value)
    assert str(validate.minimum) in str(e.value)


def test_size_minimum():
    validate = _options.Size(minimum=42)
    assert validate("42") == 42
    with pytest.raises(ValueError) as e:
        validate("41")
    assert repr("41") in str(e.value)
    assert str(validate.minimum) in str(e.value)


def test_size_minimum_none():
    with pytest.raises(TypeError):
        _options.Size(minimum=None)


def test_size_maximum():
    validate = _options.Size(maximum=42)
    assert validate("42") == 42
    with pytest.raises(ValueError) as e:
        validate("43")
    assert repr("43") in str(e.value)
    assert str(validate.maximum) in str(e.value)


def test_size_values():
    validate = _options.Size(
        minimum=4 * KiB,
        default=2 * MiB,
        maximum=1 * GiB)
    assert validate.minimum == 4 * KiB
    assert str(validate.minimum) == "4k"
    assert validate.default == 2 * MiB
    assert str(validate.default) == "2m"
    assert validate.maximum == 1 * GiB
    assert str(validate.maximum) == "1g"


def test_uuid_valid():
    validate = _options.UUID
    id = str(uuid.uuid4())
    assert validate(id) == id


def test_uuid_invalid():
    validate = _options.UUID
    id = "invalid uuid"
    with pytest.raises(ValueError) as e:
        validate(id)
    assert id in str(e.value)
    assert "UUID" in str(e.value)


@pytest.fixture
def config(tmpdir, monkeypatch):
    monkeypatch.setitem(os.environ, "XDG_CONFIG_HOME", str(tmpdir))
    config = tmpdir.join("ovirt-img.conf")
    config.write("""
[all]
engine_url = https://engine.com
username = username
password = password
cafile = /engine.pem
secure = False
disk_timeout = 200
log_file = /var/log/ovirt-img/engine.log
log_level = info

[required]
engine_url = https://engine.com
username = username

[missing1]
engine_url = https://engine.com
#username is not set

[missing2]
#engine_url is not set
username = username

[missing3]
#engine_url is not set
#username is not set
password = password

[invalid_log_level]
engine_url = https://engine.com
username = username
password = password
log_level = invalid value

[invalid_secure]
engine_url = https://engine.com
username = username
password = password
secure = invalid value
""")


def test_config_all(config):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    args = parser.parse(["test", "-c", "all"])
    assert args.engine_url == "https://engine.com"
    assert args.username == "username"
    assert args.cafile == "/engine.pem"
    assert args.secure is False
    assert args.output == "text"
    assert args.disk_timeout == 200
    assert args.log_file == "/var/log/ovirt-img/engine.log"
    assert args.log_level == "info"
    assert args.max_workers == 4
    assert args.buffer_size == 4 * MiB

    # Use password from config.
    assert args.password_file is None
    assert args.password == "password"


def test_config_all_override(config, tmpdir):
    password_file = tmpdir.join("password")
    password_file.write("password")

    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    args = parser.parse([
        "test",
        "-c", "all",
        "--engine-url", "https://engine2.com",
        "--username", "username2",
        "--password-file", str(password_file),
        "--cafile", "/engine2.pem",
        "--log-file", "test.log",
        "--log-level", "debug",
    ])
    assert args.engine_url == "https://engine2.com"
    assert args.username == "username2"
    assert args.cafile == "/engine2.pem"
    assert args.secure is False
    assert args.output == "text"
    assert args.disk_timeout == 200
    assert args.log_file == "test.log"
    assert args.log_level == "debug"
    assert args.max_workers == 4
    assert args.buffer_size == 4 * MiB

    # --password-file overrides password from config.
    assert args.password_file == str(password_file)
    assert args.password == "password"


def test_config_required(config, monkeypatch):
    monkeypatch.setattr(getpass, "getpass", lambda: "password")

    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    args = parser.parse(["test", "-c", "required"])
    assert args.engine_url == "https://engine.com"
    assert args.username == "username"
    assert args.cafile is None
    assert args.secure is True
    assert args.output == "text"
    assert args.disk_timeout == 120
    assert args.log_file is None
    assert args.log_level == "warning"
    assert args.max_workers == 4
    assert args.buffer_size == 4 * MiB

    # No --password-file or config password: use getpass.getpass().
    assert args.password_file is None
    assert args.password == "password"


def test_config_required_override(config, tmpdir):
    password_file = tmpdir.join("password")
    password_file.write("password")

    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    args = parser.parse([
        "test",
        "-c", "required",
        "--password-file", str(password_file),
        "--cafile", "/engine2.pem",
        "--insecure",
        "--log-file", "test.log",
        "--log-level", "debug",
    ])
    assert args.engine_url == "https://engine.com"
    assert args.username == "username"
    assert args.cafile == "/engine2.pem"
    assert args.secure is False
    assert args.output == "text"
    assert args.disk_timeout == 120
    assert args.log_file == "test.log"
    assert args.log_level == "debug"
    assert args.max_workers == 4
    assert args.buffer_size == 4 * MiB

    # Read password from --password-file.
    assert args.password_file == password_file
    assert args.password == "password"


@pytest.mark.parametrize("name,missing", [
    ("missing1", ["username"]),
    ("missing2", ["engine_url"]),
    ("missing3", ["engine_url", "username"]),
])
def test_config_missing(config, capsys, name, missing):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    with pytest.raises(SystemExit):
        parser.parse(["test", "-c", name])
    captured = capsys.readouterr()
    for name in missing:
        assert name in captured.err


def test_config_missing_override(config):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    args = parser.parse([
        "test",
        "-c", "missing3",
        "--engine-url", "https://engine.com",
        "--username", "username",
    ])
    assert args.engine_url == "https://engine.com"
    assert args.username == "username"


@pytest.mark.parametrize("name", [
    "invalid_log_level",
    "invalid_secure",
])
def test_config_invalid(config, capsys, name):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    with pytest.raises(SystemExit):
        parser.parse(["test", "-c", name])
    captured = capsys.readouterr()
    assert repr("invalid value") in captured.err


def test_config_no_section(config, capsys):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)

    # Required paremeters are available via the command line, but the specified
    # configuration does not exist. This is likely a user error so fail loudly.
    with pytest.raises(SystemExit):
        parser.parse([
            "test",
            "-c", "nosection",
            "--engine-url", "https://engine.com",
            "--username", "username",
        ])

    captured = capsys.readouterr()
    assert repr("nosection") in captured.err


def test_transfer_options(config):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)

    args = parser.parse([
        "test",
        "-c", "all",
        "--max-workers", "8",
        "--buffer-size", "16m"
    ])
    assert args.max_workers == 8
    assert args.buffer_size == 16 * MiB


def test_transfer_options_disabled(config):
    parser = _options.Parser()
    parser.add_sub_command(
        "test", "help", lambda x: None, transfer_options=False)
    args = parser.parse([
        "test",
        "-c", "all",
    ])
    assert not hasattr(args, 'max_workers')
    assert not hasattr(args, 'buffer_size')


@pytest.mark.parametrize("output", [
    "text",
    "json",
])
def test_output_option(config, output):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)

    args = parser.parse([
        "test",
        "-c", "all",
        "--output", output
    ])
    assert args.output == output


def test_invalid_output_option(config, capsys):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)

    with pytest.raises(SystemExit):
        parser.parse([
            "test",
            "-c", "all",
            "--output", "invalid value"
        ])
    captured = capsys.readouterr()
    assert repr("invalid value") in captured.err


def test_auto_help(capsys):
    # Running without arguments is same as running with --help.
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)

    with pytest.raises(SystemExit):
        parser.parse([])
    err1 = capsys.readouterr().err

    with pytest.raises(SystemExit):
        parser.parse(["--help"])
    err2 = capsys.readouterr().err

    assert err1 == err2


def test_version(capsys):
    parser = _options.Parser()
    parser.add_sub_command("test", "help", lambda x: None)
    with pytest.raises(SystemExit):
        parser.parse(["--version"])
    out = capsys.readouterr().out

    assert version.string in out



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import configparser
import glob
import grp
import json
import logging
import logging.config
import os
import pwd
import signal
import socket
import sys

from . import auth
from . import config
from . import errors
from . import services
from . import version

DEFAULT_CONF_DIR = "/etc/ovirt-imageio"
VENDOR_CONF_DIR = "/usr/lib/ovirt-imageio"

log = logging.getLogger("server")


def main():
    args = parse_args()
    try:
        cfg = load_config(args.conf_dir)
        if args.show_config:
            show_config(cfg)
            return

        configure_logger(cfg)
        log.info("Starting (hostname=%s pid=%s, version=%s)",
                 socket.gethostname(), os.getpid(), version.string)

        server = Server(cfg, ticket=args.ticket)
        signal.signal(signal.SIGINT, server.terminate)
        signal.signal(signal.SIGTERM, server.terminate)

        server.start()
        try:
            notify_systemd(cfg)
            log.info("Ready for requests")
            while server.running:
                signal.pause()
        finally:
            server.stop()
        log.info("Stopped")
    except Exception:
        log.exception("Server failed")
        sys.exit(1)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--conf-dir",
        default=DEFAULT_CONF_DIR,
        help="path to configuration directory, where daemon.conf and "
             "logger.conf are located.")
    parser.add_argument(
        "--show-config",
        action="store_true",
        help="print actual configuration in json format and exit. This is "
             "useful for debugging configuration issues, or reading imageio "
             "configuration by other programs.")
    parser.add_argument(
        "-t", "--ticket",
        help="path to a ticket to load during server startup.")
    return parser.parse_args()


def show_config(cfg):
    print(json.dumps(config.to_dict(cfg), indent=4))


def load_config(conf_dir):
    pattern = os.path.join(conf_dir, "conf.d", "*.conf")
    files = glob.glob(pattern)

    # Trying to use invalid configuration directory is a user error or
    # broken installation. Failing fast will help to fix the issue.
    # https://github.com/oVirt/ovirt-imageio/issues/33
    if not files:
        raise ValueError(f"Could not find {pattern}")

    # Vendor may override application defaults if needed.
    pattern = os.path.join(VENDOR_CONF_DIR, "conf.d", "*.conf")
    files.extend(glob.glob(pattern))

    # Override files based on file name sort order:
    # - /var/lib/ovirt-imageio/conf.d/75-vendor.conf overrides
    #   /etc/ovirt-imageio/conf.d/50-vdsm.conf.
    # - /etc/ovirt-imageio/conf.d/99-user.conf overrides
    #   /var/lib/ovirt-imageio/conf.d/75-vendor.conf.
    files.sort(key=os.path.basename)

    return config.load(files)


def configure_logger(cfg):
    parser = configparser.RawConfigParser()
    parser.read_dict(config.to_dict(cfg))
    logging.config.fileConfig(parser, disable_existing_loggers=False)


def notify_systemd(cfg):
    if cfg.daemon.systemd_enable:
        log.debug("Notifying systemd")
        import systemd.daemon
        systemd.daemon.notify("READY=1")


class Server:

    def __init__(self, config, ticket=None):
        self.config = config
        self.running = False
        self.auth = auth.Authorizer(config)
        self.remote_service = services.RemoteService(self.config, self.auth)

        self.local_service = None
        if config.local.enable:
            self.local_service = services.LocalService(self.config, self.auth)

        self.control_service = None
        if config.control.enable:
            self.control_service = services.ControlService(
                self.config, self.auth)

        if os.geteuid() == 0 and self.config.daemon.drop_privileges:
            self._drop_privileges()

        if ticket:
            self._add_ticket(ticket)

    def start(self):
        assert not self.running
        self.running = True

        self.remote_service.start()
        if self.local_service is not None:
            self.local_service.start()
        if self.control_service is not None:
            self.control_service.start()

    def stop(self):
        log.debug("Stopping services")
        self.remote_service.stop()
        if self.local_service is not None:
            self.local_service.stop()
        if self.control_service is not None:
            self.control_service.stop()

    def terminate(self, signo, frame):
        log.info("Received signal %d, shutting down", signo)
        self.running = False

    def _drop_privileges(self):
        uid = pwd.getpwnam(self.config.daemon.user_name).pw_uid
        gid = grp.getgrnam(self.config.daemon.group_name).gr_gid

        # Restore ownership of run directory.
        run_dir = self.config.daemon.run_dir
        if os.path.exists(run_dir):
            log.debug(
                "Changing ownership of %s to %i:%i" % (run_dir, uid, gid))
            os.chown(run_dir, uid, gid)

        # Restore ownership of log files.
        for h in logging.root.handlers:
            # Support only logging.FileHandler and sub-classes.
            filename = getattr(h, "baseFilename", None)
            if filename is not None:
                log.debug(
                    "Changing ownership of %s to %i:%i" % (filename, uid, gid))
                os.chown(filename, uid, gid)

        # Restore ownership of control socket if used.
        if self.control_service is not None:
            transport = self.config.control.transport.lower()
            if transport == "unix":
                os.chown(self.config.control.socket, uid, gid)

        # Set new uid and gid for the process.
        log.debug("Dropping root privileges, running as %i:%i" % (uid, gid))
        os.initgroups(self.config.daemon.user_name, gid)
        os.setgid(gid)
        os.setuid(uid)

    def _add_ticket(self, ticket):
        log.info("Initial ticket: %s", ticket)
        self.auth.add(self._read_ticket(ticket))

    def _read_ticket(self, ticket):
        try:
            with open(ticket, 'r') as f:
                return json.loads(f.read())
        except ValueError as e:
            raise errors.InvalidTicket(
                f"Cannot parse ticket {ticket}: {e}") from e
        except FileNotFoundError as e:
            raise errors.InvalidTicket(f"Cannot read ticket: {e}") from e



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import io
import os
import subprocess
import urllib.parse

from contextlib import closing

import pytest
import userstorage

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import extent
from ovirt_imageio._internal import util
from ovirt_imageio._internal.backends import file

from .. import storage

BACKENDS = userstorage.load_config("storage.py").BACKENDS

# Up to kernel 5.5, block size detection for unallocated file was broken only
# on XFS.  Since 5.5 it is broken on all file systems.
KERNEL_VERSION = tuple(map(int, os.uname()[2].split(".")[:2]))


@pytest.fixture(
    params=[
        # backend, can_detect_sector_size.
        (BACKENDS["file-512-ext2"], KERNEL_VERSION < (5, 5)),
        (BACKENDS["file-512-ext4"], KERNEL_VERSION < (5, 5)),
        (BACKENDS["file-512-xfs"], False),
        (BACKENDS["file-4k-ext2"], True),
        (BACKENDS["file-4k-ext4"], True),
        (BACKENDS["file-4k-xfs"], True),
    ],
    ids=lambda x: str(x[0]),
)
def user_file(request):
    backend, can_detect_sector_size = request.param
    with backend:
        yield storage.Backend(backend, can_detect_sector_size)


def test_debugging_interface(user_file):
    with file.open(user_file.url, "r+") as f:
        assert f.readable()
        assert f.writable()
        assert not f.sparse
        assert f.name == "file"


def test_open_read_only(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * user_file.sector_size)
    with file.open(user_file.url) as f, \
            closing(util.aligned_buffer(user_file.sector_size)) as buf:
        assert f.readable()
        assert not f.writable()
        f.readinto(buf)
        assert buf[:] == b"x" * user_file.sector_size


def test_open_read_write(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * user_file.sector_size)
    with file.open(user_file.url, "r+") as f, \
            closing(util.aligned_buffer(user_file.sector_size)) as buf:
        assert f.readable()
        assert f.writable()
        f.readinto(buf)
        buf[:] = b"b" * user_file.sector_size
        f.seek(0)
        f.write(buf)
    with io.open(user_file.path, "rb") as f:
        assert f.read() == b"b" * user_file.sector_size


@pytest.mark.parametrize("mode", ["r", "r+"])
def test_open_no_create(mode):
    with pytest.raises(OSError) as e:
        missing = urllib.parse.urlparse("file:/no/such/path")
        with file.open(missing, mode):
            pass
    assert e.value.errno == errno.ENOENT


def test_close(tmpurl):
    with file.open(tmpurl, "r+") as b:
        pass

    # Closing twice does nothing.
    b.close()

    # But other oprations should fail.
    error = "Operation on closed backend"

    with closing(util.aligned_buffer(4096)) as buf:
        with pytest.raises(ValueError) as e:
            b.write(buf)
        assert str(e.value) == error

        with pytest.raises(ValueError) as e:
            b.seek(0)
        assert str(e.value) == error

        with pytest.raises(ValueError):
            b.readinto(buf)
        assert str(e.value) == error


@pytest.mark.parametrize("size", [511, 4097])
def test_block_size_sparse(user_file, size):
    with io.open(user_file.path, "wb") as f:
        f.truncate(size)

    with file.open(user_file.url) as f:
        if user_file.can_detect_sector_size:
            assert f.block_size == user_file.sector_size
        else:
            assert f.block_size in (user_file.sector_size, 4096)

    with io.open(user_file.path, "rb") as f:
        assert f.read(size) == b"\0" * size


@pytest.mark.parametrize("size", [511, 4097])
def test_block_size_preallocated(user_file, size):
    # This is how vdsm preallocates volumes. This uses fallocate() or fallback
    # to writing one byte per block.
    subprocess.check_output(
        ["fallocate", "--posix", "--length", str(size), user_file.path])

    with file.open(user_file.url) as f:
        if user_file.can_detect_sector_size:
            assert f.block_size == user_file.sector_size
        else:
            assert f.block_size in (user_file.sector_size, 4096)

    with io.open(user_file.path, "rb") as f:
        assert f.read(size) == b"\0" * size


def test_block_size_first_block_unallocated(user_file):
    # Assumes file system block size <= 4096.
    hole_size = 4096
    data_size = 4096

    with io.open(user_file.path, "wb") as f:
        f.seek(hole_size)
        f.write(b"x" * data_size)

    with file.open(user_file.url) as f:
        if user_file.can_detect_sector_size:
            assert f.block_size == user_file.sector_size
        else:
            assert f.block_size in (user_file.sector_size, 4096)

    with io.open(user_file.path, "rb") as f:
        assert f.read(hole_size) == b"\0" * hole_size
        assert f.read(data_size) == b"x" * data_size


@pytest.mark.parametrize("size", [511, 4097])
def test_block_size_allocated(user_file, size):
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    with file.open(user_file.url) as f:
        assert f.block_size == user_file.sector_size

    with io.open(user_file.path, "rb") as f:
        assert f.read() == b"x" * size


def test_readinto(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * 4096)
    with file.open(user_file.url) as f, \
            closing(util.aligned_buffer(4096)) as buf:
        n = f.readinto(buf)
        assert n == len(buf)
        assert f.tell() == len(buf)
        assert buf[:] == b"a" * 4096


def test_readinto_short_aligned(user_file):
    size = user_file.sector_size
    buf_size = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * size)

    with file.open(user_file.url) as f, \
            closing(util.aligned_buffer(buf_size)) as buf:
        n = f.readinto(buf)
        assert n == size
        assert f.tell() == size
        assert buf[:size] == b"a" * size
        assert buf[size:] == b"\0" * (buf_size - size)


def test_readinto_short_unaligned(user_file):
    size = 42
    buf_size = user_file.sector_size

    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * size)

    with file.open(user_file.url) as f, \
            closing(util.aligned_buffer(buf_size)) as buf:
        n = f.readinto(buf)
        assert n == size
        assert f.tell() == size
        assert buf[:size] == b"a" * size
        assert buf[size:] == b"\0" * (buf_size - size)


def test_write_aligned_middle(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * 4 * 4096)
    with file.open(user_file.url, "r+") as f, \
            closing(util.aligned_buffer(8192)) as buf:
        buf.write(b"b" * 8192)
        f.seek(4096)
        n = f.write(buf)
        assert n == len(buf)
        assert f.tell() == 4096 + len(buf)
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"a" * 4096
        assert f.read(8192) == b"b" * 8192
        assert f.read() == b"a" * 4096


def test_write_aligned_at_end(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * 8192)
    with file.open(user_file.url, "r+") as f, \
            closing(util.aligned_buffer(8192)) as buf:
        buf.write(b"b" * 8192)
        f.seek(4096)
        n = f.write(buf)
        assert n == len(buf)
        assert f.tell() == 4096 + len(buf)
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"a" * 4096
        assert f.read() == b"b" * 8192


def test_write_aligned_after_end(user_file):
    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * 4096)
    with file.open(user_file.url, "r+") as f, \
            closing(util.aligned_buffer(8192)) as buf:
        buf.write(b"b" * 8192)
        f.seek(8192)
        n = f.write(buf)
        assert n == len(buf)
        assert f.tell() == 8192 + len(buf)
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"a" * 4096
        assert f.read(4096) == b"\0" * 4096
        assert f.read() == b"b" * 8192


def test_write_unaligned_offset_complete(user_file):
    size = user_file.sector_size * 2
    start = user_file.sector_size + 10
    end = start + 10

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Write 10 bytes into the second block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.write(b"y" * 10)
        assert n == 10
        assert f.tell() == end

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(10) == b"y" * 10
        assert f.read() == b"x" * (size - end)


def test_write_unaligned_offset_inside(user_file):
    size = user_file.sector_size * 2
    start = user_file.sector_size - 12

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Write 12 bytes into the first block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.write(b"y" * 100)
        assert n == 12
        assert f.tell() == user_file.sector_size

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(12) == b"y" * 12
        assert f.read() == b"x" * user_file.sector_size


def test_write_unaligned_offset_at_end(user_file):
    size = user_file.sector_size * 2
    start = size - 24

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Write 24 bytes into the last block.
    with file.open(user_file.url, "r+") as f:
        f.seek(size - 24)
        n = f.write(b"y" * 100)
        assert n == 24
        assert f.tell() == size

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read() == b"y" * 24


def test_write_unaligned_offset_after_end(user_file):
    size = user_file.sector_size

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    with file.open(user_file.url, "r+") as f:
        f.seek(size + 10)
        n = f.write(b"y" * 10)
        assert n == 10
        assert f.tell() == size + 20

    with io.open(user_file.path, "rb") as f:
        assert f.read(size) == b"x" * size
        assert f.read(10) == b"\0" * 10
        assert f.read(10) == b"y" * 10
        assert f.read() == b"\0" * (user_file.sector_size - 20)


def test_write_unaligned_buffer_slow_path(user_file):
    size = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Perform slow read-modify-write in the second block.
    with file.open(user_file.url, "r+") as f:
        f.seek(user_file.sector_size)
        n = f.write(b"y" * 10)
        assert n == 10
        assert f.tell() == user_file.sector_size + 10

    with io.open(user_file.path, "rb") as f:
        assert f.read(user_file.sector_size) == b"x" * user_file.sector_size
        assert f.read(10) == b"y" * 10
        assert f.read() == b"x" * (user_file.sector_size - 10)


def test_write_unaligned_buffer_fast_path(user_file):
    size = user_file.sector_size * 4
    start = user_file.sector_size
    length = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Create buffer of 2 blcoks + 1 byte.
    buf = util.aligned_buffer(length + 1)
    buf.write(b"y" * len(buf))

    # Perform short fast write of 2 blocks.
    with closing(buf):
        with file.open(user_file.url, "r+") as f:
            f.seek(start)
            n = f.write(buf)
            assert n == length
            assert f.tell() == start + length

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(length) == b"y" * length
        assert f.read() == b"x" * user_file.sector_size


def test_flush(user_file, monkeypatch):
    count = [0]

    def fsync(fd):
        count[0] += 1

    # This is ugly but probably the only way to test that we call fsync.
    monkeypatch.setattr(os, "fsync", fsync)
    with file.open(user_file.url, "r+") as f:
        f.write(b"x")
        f.flush()
    assert count[0] == 1


ZERO_SPARSE = [
    pytest.param(True, id="sparse"),
    pytest.param(False, id="preallocted"),
]


@pytest.mark.parametrize("sparse", ZERO_SPARSE)
def test_zero_aligned_middle(user_file, sparse):
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * 3 * 4096)
    with file.open(user_file.url, "r+", sparse=sparse) as f:
        f.seek(4096)
        n = f.zero(4096)
        assert n == 4096
        assert f.tell() == 8192
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"x" * 4096
        assert f.read(4096) == b"\0" * 4096
        assert f.read() == b"x" * 4096


@pytest.mark.parametrize("sparse", ZERO_SPARSE)
def test_zero_aligned_at_end(user_file, sparse):
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * 4096)
    with file.open(user_file.url, "r+", sparse=sparse) as f:
        assert f.sparse == sparse
        f.seek(4096)
        n = f.zero(4096)
        assert n == 4096
        assert f.tell() == 8192
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"x" * 4096
        assert f.read() == b"\0" * 4096


@pytest.mark.parametrize("sparse", ZERO_SPARSE)
def test_zero_aligned_after_end(user_file, sparse):
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * 4096)
    with file.open(user_file.url, "r+", sparse=sparse) as f:
        f.seek(8192)
        n = f.zero(4096)
        assert n == 4096
        assert f.tell() == 12288
    with io.open(user_file.path, "rb") as f:
        assert f.read(4096) == b"x" * 4096
        assert f.read() == b"\0" * 8192


def test_zero_allocate_space(user_file):
    size = 1024**2
    with file.open(user_file.url, "r+", sparse=False) as f:
        f.zero(size)
    # File system may report more than file size.
    assert os.stat(user_file.path).st_blocks * 512 >= size


def test_zero_sparse_deallocate_space(user_file):
    size = 1024**2
    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)
    with file.open(user_file.url, "r+", sparse=True) as f:
        f.zero(size)
    assert os.stat(user_file.path).st_blocks * 512 < size


def test_zero_unaligned_offset_complete(user_file):
    size = user_file.sector_size * 2
    start = user_file.sector_size + 10

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Zero 10 bytes into the second block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(10)
        assert n == 10
        assert f.tell() == start + 10

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(10) == b"\0" * 10
        assert f.read() == b"x" * (size - start - 10)


def test_zero_unaligned_offset_inside(user_file):
    size = user_file.sector_size * 2
    start = user_file.sector_size - 10

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Zero 10 bytes into the first block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(100)
        assert n == 10
        assert f.tell() == start + 10

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(10) == b"\0" * 10
        assert f.read() == b"x" * (size - start - 10)


def test_zero_unaligned_offset_at_end(user_file):
    size = user_file.sector_size * 2
    start = size - 10

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Zero 10 bytes into the last block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(100)
        assert n == 10
        assert f.tell() == size

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read() == b"\0" * 10


def test_zero_unaligned_offset_after_end(user_file):
    size = user_file.sector_size
    start = size + 10

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(10)
        assert n == 10
        assert f.tell() == start + 10

    with io.open(user_file.path, "rb") as f:
        assert f.read(size) == b"x" * size
        assert f.read() == b"\0" * user_file.sector_size


def test_zero_unaligned_buffer_slow_path(user_file):
    size = user_file.sector_size * 2
    start = user_file.sector_size

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Perform slow read-modify-write in the second block.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(10)
        assert n == 10
        assert f.tell() == start + 10

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(10) == b"\0" * 10
        assert f.read() == b"x" * (size - start - 10)


def test_zero_unaligned_buffer_fast_path(user_file):
    size = user_file.sector_size * 4
    start = user_file.sector_size
    length = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    # Perform fast short zero of 2 blocks.
    with file.open(user_file.url, "r+") as f:
        f.seek(start)
        n = f.zero(length + 1)
        assert n == length
        assert f.tell() == start + length

    with io.open(user_file.path, "rb") as f:
        assert f.read(start) == b"x" * start
        assert f.read(length) == b"\0" * length
        assert f.read() == b"x" * (size - start - length)


def test_dirty(user_file):
    # backend created clean
    with file.open(user_file.url, "r+", sparse=True) as f:
        assert not f.dirty
        buf = util.aligned_buffer(4096)
        with closing(buf):
            # write ans zero dirty the backend
            buf.write(b"x" * 4096)
            f.write(buf)
            assert f.dirty
            f.flush()
            assert not f.dirty
            f.zero(4096)
            assert f.dirty
            f.flush()
            assert not f.dirty

            # readinto, seek do not affect dirty.
            f.seek(0)
            assert not f.dirty
            f.readinto(buf)
            assert not f.dirty


def test_size(user_file):
    size = 4096

    with io.open(user_file.path, "wb") as f:
        f.truncate(size)

    with file.open(user_file.url, "r+", sparse=True) as f:
        # Check initial size.
        f.seek(100)
        assert f.size() == size
        assert f.tell() == 100

        # Check that size() updates when file size is modified.
        f.seek(size)
        f.zero(4096)
        assert f.size() == 8192


def test_extents(user_file):
    size = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.truncate(size)

    with file.open(user_file.url, "r+", sparse=True) as f:
        # We support detecting extents now; empty file reports one data
        # extents.
        assert list(f.extents()) == [
            extent.ZeroExtent(0, size, False, False)
        ]


def test_extents_dirty(user_file):
    with file.open(user_file.url, "r+", dirty=True) as f:
        with pytest.raises(errors.UnsupportedOperation):
            list(f.extents(context="dirty"))


def test_clone(user_file):
    size = user_file.sector_size * 2

    with io.open(user_file.path, "wb") as f:
        f.write(b"x" * size)

    with file.open(user_file.url, "r+", sparse=True) as a, \
            a.clone() as b, \
            util.aligned_buffer(user_file.sector_size) as buf:

        # Backends are identical when created.
        assert a.size() == b.size()
        assert a.tell() == b.tell()
        assert a.block_size == b.block_size

        # Operations on one backend do not affect the other.
        buf[:] = b"y" * len(buf)
        a.write(buf)
        assert a.tell() == len(buf)
        assert b.tell() == 0

        # But both expose the same file contents.
        buf[:] = b"\0" * len(buf)
        b.readinto(buf)
        assert buf[:] == b"y" * len(buf)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import os

from .. import errors
from .. import extent
from .. import nbd
from .. import nbdutil

from . common import CLOSED

log = logging.getLogger("backends.nbd")

Error = nbd.Error


def open(url, mode="r", sparse=False, dirty=False, max_connections=8,
         **options):
    """
    Open a NBD backend.

    Arguments:
        url (url): parsed NBD url (e.g. "nbd:unix:/socket:exportname=name")
        mode (str): "r" for readonly, "w" for write only, "r+" for read write.
            Note that writeable backend does not guarantee that the underling
            nbd server is writable. The server must be configured to allow
            writing.
        sparse (bool): deallocate space when zeroing if possible.
        dirty (bool): if True, configure the client to report dirty extents.
            Can work only when connecting to qemu during incremental backup.
        max_connections (int): maximum number of connections per backend
            allowed on this server. Limit backends's max_readers and
            max_writers.
        **options: ignored, nbd backend does not have any options.
    """
    client = nbd.open(url, dirty=dirty)
    try:
        return Backend(
            client,
            mode=mode,
            sparse=sparse,
            max_connections=max_connections)
    except:  # noqa: E722
        client.close()
        raise


class Backend:
    """
    NBD backend.
    """

    def __init__(self, client, mode="r", sparse=False, max_connections=8):
        if mode not in ("r", "w", "r+"):
            raise ValueError("Unsupported mode %r" % mode)
        log.debug("Open address=%r export_name=%r sparse=%r "
                  "max_connections=%r",
                  client.address, client.export_name, sparse, max_connections)
        self._client = client
        self._mode = mode
        self._sparse = sparse
        self._position = 0
        self._dirty = False
        self._max_connections = max_connections

    def clone(self):
        """
        Return new backend connected to the same NBD export.
        """
        client = nbd.Client(
            self._client.address,
            export_name=self._client.export_name,
            dirty=self._client.dirty)
        try:
            return self.__class__(
                client,
                mode=self._mode,
                max_connections=self._max_connections)
        except:  # noqa: E722
            client.close()
            raise

    @property
    def max_readers(self):
        return self._max_connections

    @property
    def max_writers(self):
        return self._max_connections

    # Backend interface

    def readinto(self, buf):
        if not self.readable():
            raise IOError("Unsupported operation: readinto")

        length = min(len(buf), self._client.export_size - self._position)
        if length <= 0:
            # Client should not send NBD_CMD_READ with zero length, and
            # request after the end of file is invalid.
            return 0

        with memoryview(buf)[:length] as view:
            self._client.readinto(self._position, view)

        self._position += length
        return length

    def write(self, buf):
        if not self.writable():
            raise IOError("Unsupported operation: write")
        self._client.write(self._position, buf)
        length = len(buf)
        self._position += length
        self._dirty = True
        return length

    def zero(self, length):
        if not self.writable():
            raise IOError("Unsupported operation: zero")
        self._client.zero(self._position, length, punch_hole=self._sparse)
        self._position += length
        self._dirty = True
        return length

    def flush(self):
        self._client.flush()
        self._dirty = False

    def tell(self):
        return self._position

    def seek(self, pos, how=os.SEEK_SET):
        if how == os.SEEK_SET:
            self._position = pos
        elif how == os.SEEK_CUR:
            self._position += pos
        elif how == os.SEEK_END:
            self._position = self._client.export_size + pos
        return self._position

    def close(self):
        if self._client is not CLOSED:
            log.debug("Close address=%r", self._client.address)
            self._client.close()
            self._client = CLOSED

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            # Do not hide the original error.
            if t is None:
                raise
            log.exception("Error closing")

    def extents(self, context="zero"):
        if context not in ("zero", "dirty"):
            raise errors.UnsupportedOperation(
                "Backend nbd does not support {} extents".format(context))

        # If server does not support base:allocation, we can safely report one
        # data extent like other backends.
        if context == "zero" and not self._client.has_base_allocation:
            yield extent.ZeroExtent(0, self._client.export_size, False, False)
            return

        # If dirty extents are not available, client may be able to use zero
        # extents for eficient download, so we should not fake the response.
        if context == "dirty" and self._client.dirty_bitmap is None:
            raise errors.UnsupportedOperation(
                "NBD export {!r} does not support dirty extents"
                .format(self._client.export_name))

        dirty = context == "dirty"
        start = 0
        for ext in nbdutil.extents(self._client, dirty=dirty):
            if dirty:
                yield extent.DirtyExtent(
                    start, ext.length, ext.dirty, ext.zero)
            else:
                yield extent.ZeroExtent(
                    start, ext.length, ext.zero, ext.hole)
            start += ext.length

    @property
    def block_size(self):
        # qemu always reports minium_block_size=1, so caller never needs to
        # align requests and read more data than needed.
        return self._client.minimum_block_size

    # Debugging interface

    def readable(self):
        return self._mode in ("r", "r+")

    def writable(self):
        return self._mode in ("w", "r+")

    @property
    def dirty(self):
        """
        Returns True if backend was modified and needs flushing.
        """
        return self._dirty

    @property
    def sparse(self):
        return self._sparse

    @property
    def name(self):
        return "nbd"

    def size(self):
        return self._client.export_size



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Helpers for working with oVirt python SDK.
"""

import logging
import time

import ovirtsdk4 as sdk
from ovirtsdk4 import types

from . _options import ADD_DISK_TIMEOUT

log = logging.getLogger("ovirt")

# Image transfer constants.
DOWNLOAD = types.ImageTransferDirection.DOWNLOAD
RAW = types.DiskFormat.RAW
COW = types.DiskFormat.COW
ISO = types.DiskContentType.ISO
DATA = types.DiskContentType.DATA


class Repr:
    """
    Helper for logging sdk objects in a useful way.
    """

    def __init__(self, obj, key="id"):
        self.obj = obj
        self.key = key

    def __repr__(self):
        if self.obj is None:
            return "None"

        typename = type(self.obj).__name__
        value = getattr(self.obj, self.key, None)
        return f"{typename}({self.key}={value!r})"


def connect(args):
    return sdk.Connection(
        url=f"{args.engine_url}/ovirt-engine/api",
        username=args.username,
        password=args.password,
        ca_file=args.cafile,
        insecure=not args.secure,
        log=log,
        debug=args.log_level == "debug")


def find_disk(con, disk_id):
    service = _disk_service(con, disk_id)
    return service.get()


def add_disk(con, name, provisioned_size, sd_name, id=None,
             initial_size=None, sparse=True, enable_backup=True,
             content_type=DATA, format=COW, timeout=ADD_DISK_TIMEOUT):
    """
    Add a new disk to the storage domain, based on the source image
    information provided.

    Arguments:
        con (ovirtsdk4.Connection): connection to oVirt engine
        name (str): New disk alias.
        provisioned_size (int): Provisioned size of the new disk.
        sd_name (str): Storage Domain name that will contain the new disk.
        id (str): ID of the new disk to be added. By default oVirt creates
            a new UUID when not specified.
        initial_size (int): Initial size of the new disk.
        sparse (bool): New disk is sparse.
        enable_backup (bool): Disk can be used for incremental backups.
        content_type (ovirtsdk4.types.DiskContentType): Content type for the
            new disk.
        format (ovirtsdk4.types.DiskFormat): Format of the new disk.
        timeout (int, optional): number of seconds to wait for
            disk to be created.

    Returns:
        ovirtsdk4.types.Disk
    """
    log.info(
        "Adding disk name=%s provisioned_size=%s sd_name=%s id=%s "
        "initial_size=%s sparse=%s enable_backup=%s, content_type=%s "
        "format=%s", name, provisioned_size, sd_name, id, initial_size,
        sparse, enable_backup, content_type, format)

    disks_service = con.system_service().disks_service()
    disk = disks_service.add(
        disk=types.Disk(
            id=id,
            name=name,
            content_type=content_type,
            description='Uploaded by ovirt-img',
            format=format,
            initial_size=initial_size,
            provisioned_size=provisioned_size,
            sparse=sparse,
            backup=types.DiskBackup.INCREMENTAL if enable_backup else None,
            storage_domains=[
                types.StorageDomain(
                    name=sd_name
                )
            ]
        )
    )
    _wait_for_disk(con, disk.id, timeout)
    return disk


def find_storage_domain(con, disk):
    service = _storage_domain_service(con, disk.storage_domains[0].id)
    return service.get()


def find_host(con, sd_name):
    """
    Check if we can perform a transfer using the local host and return a
    host instance. Return None if we cannot use this host.

    Using the local host for an image transfer allows optimizing the
    connection using unix socket. This speeds up the transfer
    significantly and minimizes the network bandwidth.

    However using the local host is possible only if:
    - The local host is an oVirt host
    - The host is Up
    - The host is in the same DC of the storage domain

    Consider this setup:

        laptop1

        dc1
            host1 (down)
            host2 (up)
            sd1
                disk1

        dc2
            host3 (up)
            sd2
                disk2

    - If we run on laptop1 we cannot use the local host for any
      transfer.
    - If we run on host1, we cannot use the local host for any transfer.
    - If we run on host2, we can use use host2 for transferring disk1.
    - If we run on host3, we can use host3 for transferring disk2.

    Arguments:
        con (ovirtsdk4.Connection): Connection to ovirt engine
        sd_name (str): Storage domain name

    Returns:
        ovirtsdk4.types.Host
    """

    # Try to read this host hardware id.

    try:
        with open("/etc/vdsm/vdsm.id") as f:
            vdsm_id = f.readline().strip()
    except FileNotFoundError:
        log.debug("Not running on oVirt host, using any host")
        return None
    except OSError as e:
        # Unexpected error when running on ovirt host. Since choosing a
        # host is an optimization, log and continue.
        log.warning("Cannot read /etc/vdsm/vdsm.id, using any host: %s", e)
        return None

    log.debug("Found host hardware id: %s", vdsm_id)

    # Find the data center by storage domain name.

    system_service = con.system_service()
    data_centers = system_service.data_centers_service().list(
        search=f'storage.name={sd_name}',
        case_sensitive=True)
    if len(data_centers) == 0:
        raise RuntimeError(
            f"Storage domain {sd_name} is not attached to a DC")

    data_center = data_centers[0]
    log.debug("Found data center: %s", data_center.name)

    # Validate that this host is up and in data center.

    hosts_service = system_service.hosts_service()
    query = f"hw_id={vdsm_id} and datacenter={data_center.name} and status=Up"
    hosts = hosts_service.list(search=query, case_sensitive=True)
    if len(hosts) == 0:
        log.debug(
            "Cannot use host with hardware id %s, host is not up, or does "
            "not belong to data center %s",
            vdsm_id, data_center.name)
        return None

    host = hosts[0]
    log.debug("Using host id %s", host.id)

    return host


def create_transfer(
        con, image, direction=types.ImageTransferDirection.UPLOAD, host=None,
        backup=None, inactivity_timeout=None, timeout=60, shallow=None,
        timeout_policy=types.ImageTransferTimeoutPolicy.CANCEL):
    """
    Create an image transfer for upload or download.

    Arguments:
        con (ovirtsdk4.Connection): connection to ovirt engine
        image (ovirtsdk4.types.Disk | ovirtsdk4.types.DiskSnapshot): The image
            to transfer.
        direction (ovirtsdk4.typles.ImageTransferDirection): transfer
            direction (default UPLOAD)
        host (ovirtsdk4.types.Host): host object that should perform the
            transfer. If not specified engine will pick a random host.
        backup (ovirtsdk4.types.Backup): When downloading backup, the
            backup object owning the disks.
        inactivity_timeout (int): Number of seconds engine will wait for
            client activity before pausing the transfer. If not set, use
            engine default value.
        timeout (float, optional): number of seconds to wait for transfer
            to become ready.
        shallow (bool): Download only the specified image instead of the
            entire image chain. When downloading a disk transfer only
            the active disk snapshot data. When downloading a disk
            snapshot, transfer only the specified disk snaphost data.
        timeout_policy (ovirtsdk4.types.ImageTransferTimeoutPolicy): the
            action to take after inactivity timeout.

    Returns:
        ovirtsdk4.types.ImageTransfer in phase TRANSFERRING
    """
    log.info(
        "Creating transfer image=%s direction=%s host=%s backup=%s "
        "shallow=%s timeout_policy=%s",
        Repr(image, "id"),
        direction,
        Repr(host, "name"),
        Repr(backup, "id"),
        shallow,
        timeout_policy)

    start = time.monotonic()
    deadline = start + timeout

    transfer = types.ImageTransfer(
        host=host,
        direction=direction,
        backup=backup,
        inactivity_timeout=inactivity_timeout,
        timeout_policy=timeout_policy,
        format=RAW,
        shallow=shallow)

    if isinstance(image, types.Disk):
        transfer.disk = image
    elif isinstance(image, types.DiskSnapshot):
        transfer.snapshot = image
    else:
        raise ValueError(f"Cannot transfer {Repr(image)}")

    transfers_service = con.system_service().image_transfers_service()
    transfer = transfers_service.add(transfer)

    # At this point the transfer owns the disk and will delete the disk
    # if the transfer is canceled, or if finalizing the transfer fails.

    transfer_service = _transfer_service(con, transfer.id)

    while True:
        try:
            transfer = transfer_service.get()
        except sdk.NotFoundError:
            # The system has removed the disk and the transfer.
            raise RuntimeError(f"Transfer {transfer.id} was removed")

        if transfer.phase == types.ImageTransferPhase.FINISHED_FAILURE:
            # The system will remove the disk and the transfer soon.
            raise RuntimeError(f"Transfer {transfer.id} has failed")

        if transfer.phase == types.ImageTransferPhase.PAUSED_SYSTEM:
            transfer_service.cancel()
            raise RuntimeError(f"Transfer {transfer.id} was paused by system")

        if transfer.phase == types.ImageTransferPhase.TRANSFERRING:
            break

        if transfer.phase != types.ImageTransferPhase.INITIALIZING:
            transfer_service.cancel()
            raise RuntimeError(
                f"Unexpected transfer {transfer.id} phase {transfer.phase}")

        if time.monotonic() > deadline:
            log.info("Cancelling transfer %s", transfer.id)
            transfer_service.cancel()
            raise RuntimeError(
                f"Timed out waiting for transfer {transfer.id}")

        time.sleep(1)

    # Log the transfer host name. This is very useful for
    # troubleshooting.
    transfer.host = con.follow_link(transfer.host)

    log.info("Transfer %r ready on host %r in %.1f seconds",
             transfer.id, transfer.host.name, time.monotonic() - start)

    return transfer


def cancel_transfer(con, transfer):
    """
    Cancel a transfer and remove the disk for upload transfer.

    There is not need to cancel a download transfer, it can always be
    finalized.
    """
    log.info("Cancelling transfer %r", transfer.id)
    transfer_service = _transfer_service(con, transfer.id)
    transfer_service.cancel()


def finalize_transfer(con, transfer, disk, timeout=300):
    """
    Finalize a transfer, making the transfer disk available.

    If finalizing succeeds: the disk status will change to OK and
    transfer's phase will change to FINISHED_SUCCESS.
    On upload errors: the disk status will change to ILLEGAL, transfer's
    phase will change to FINISHED_FAILURE and the disk will be removed.
    In both cases the transfer entity:
     a. prior to 4.4.7: is removed shortly after the command finishes
     b. 4.4.7 and later: stays in the database and is cleaned by the
        dedicated thread after a few minutes.

    If oVirt fails to finalize the transfer, transfer's phase will
    change to PAUSED_SYSTEM. In this case the disk's status will change
    to ILLEGAL and it will not be removed.

    When working with oVirt 4.4.7 and later, it is enough to poll the
    image transfer. However with older versions the transfer entity is
    removed from the database after the the command finishes and before
    we can retrieve the final transfer status. Thus the API returns a
    404 error code. In that case we need to check for the disk status.

    For more info see:
    - http://ovirt.github.io/ovirt-engine-api-model/4.4/#services/image_transfer
    - http://ovirt.github.io/ovirt-engine-sdk/master/types.m.html#ovirtsdk4.types.ImageTransfer

    Arguments:
        con (ovirtsdk4.Connection): connection to ovirt engine
        transfer (ovirtsdk4.types.ImageTransfer): image transfer to
            finalize
        disk (ovirtsdk4.types.Disk): disk associated with the image
            transfer
        timeout (float, optional): number of seconds to wait for
            transfer to finalize.
    """  # noqa E501 (long line)
    log.info("Finalizing transfer %r", transfer.id)

    start = time.monotonic()

    transfer_service = _transfer_service(con, transfer.id)

    transfer_service.finalize()
    while True:
        time.sleep(1)
        try:
            transfer = transfer_service.get()
        except sdk.NotFoundError:
            # Old engine (< 4.4.7): since the transfer was already
            # deleted from the database, we can assume that the disk
            # status is already updated, so we can check it only once.
            disk_service = _disk_service(con, disk.id)
            try:
                disk = disk_service.get()
            except sdk.NotFoundError:
                # Disk verification failed and the system removed the
                # disk.
                raise RuntimeError(
                    f"Transfer {transfer.id} failed: disk {disk.id} was "
                    "removed")

            if disk.status == types.DiskStatus.OK:
                break

            raise RuntimeError(
                    f"Transfer {transfer.id} failed: disk {disk.id} status :"
                    f"{disk.status}")

        log.debug("Transfer %r in phase %r", transfer.id, transfer.phase)

        if transfer.phase == types.ImageTransferPhase.FINISHED_SUCCESS:
            break

        if transfer.phase == types.ImageTransferPhase.FINISHED_FAILURE:
            raise RuntimeError(f"Transfer {transfer.id} failed")

        if time.monotonic() > start + timeout:
            raise RuntimeError(
                f"Timed out waiting for transfer {transfer.id} to finalize, "
                f"transfer is {transfer.phase}")

    log.info("Transfer %r finalized in %.1f seconds",
             transfer.id, time.monotonic() - start)


def _wait_for_disk(con, disk_id, timeout):
    log.info("Waiting for disk %s", disk_id)
    start = time.monotonic()
    deadline = start + timeout
    disk_service = _disk_service(con, disk_id)
    while True:
        time.sleep(1)
        if time.monotonic() > deadline:
            raise RuntimeError(f"Timeout reached waiting for disk {disk_id}")

        disk = disk_service.get()
        if disk.status == types.DiskStatus.ILLEGAL:
            raise RuntimeError(f"Disk {disk_id} became ILLEGAL")

        if disk.status == types.DiskStatus.OK:
            break


def _disk_service(con, disk_id):
    return con.system_service().disks_service().disk_service(disk_id)


def _storage_domain_service(con, storage_domain_id):
    return (con.system_service()
            .storage_domains_service()
            .storage_domain_service(storage_domain_id))


def _transfer_service(con, transfer_id):
    return (con.system_service()
            .image_transfers_service()
            .image_transfer_service(transfer_id))



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

from collections import namedtuple
from functools import partial

from .. import errors
from .. import util

from . common import CLOSED
from . import file
from . import http
from . import nbd

_modules = {
    "file": file,
    "http": http,
    "https": http,
    "nbd": nbd,
}


class Unsupported(Exception):
    """ Requested backend is not supported """


class Context(namedtuple("Context", "backend,buffer")):
    """
    Backend context stored per ticket connection.
    """
    __slots__ = ()

    def close(self):
        try:
            self.backend.close()
        finally:
            self.buffer.close()


class Closer:
    """
    Call function when closed.
    """
    def __init__(self, func):
        self.close = func


class Wrapper:
    """
    Used to lend a backend without closing the wrapped backend.
    """

    def __init__(self, backend):
        self._backend = backend

    def close(self):
        self._backend = CLOSED

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()

    def __getattr__(self, name):
        return getattr(self._backend, name)


def supports(name):
    return name in _modules


def get(req, ticket, config):
    """
    Return a connection backend for this ticket.

    On the first call, open the backend and cache it in the connection context.
    The backend will be closed when the connection is closed.

    Thread safety: requests are accessed by the single connection thread, no
    locking is needed.
    """
    try:
        return ticket.get_context(req.connection_id)
    except KeyError:
        if not supports(ticket.url.scheme):
            raise Unsupported(
                "Unsupported backend {!r}".format(ticket.url.scheme))

        mode = "r+" if "write" in ticket.ops else "r"
        module = _modules[ticket.url.scheme]

        # If HTTP backend has no explict CA file configuration, use CA file
        # from TLS configuration.
        ca_file = config.backend_http.ca_file or config.tls.ca_file

        backend = module.open(
            ticket.url,
            mode=mode,
            sparse=ticket.sparse,
            dirty=ticket.dirty,
            max_connections=config.daemon.max_connections,
            cafile=ca_file)

        backend_config = getattr(config, "backend_" + backend.name)
        buf = util.aligned_buffer(backend_config.buffer_size)
        ctx = Context(backend, buf)

        # Keep the context in the ticket so we monitor the number of
        # connections using the ticket.
        try:
            ticket.add_context(req.connection_id, ctx)
        except errors.AuthorizationError:
            # The ticket was canceled. close the backend and let the caller
            # handle the failure.
            ctx.close()
            raise

        # Authorized connections get a longer timeout from the ticket.
        req.set_connection_timeout(ticket.inactivity_timeout)

        # Register a closer removing the context when the connection is closed.
        req.context[ticket.uuid] = Closer(
            partial(ticket.remove_context, req.connection_id))

        return ctx



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

def is_fedora(version=""):
    return _check("Fedora release {}".format(version))


def is_centos(version=""):
    return _check("CentOS Stream release {}".format(version))


def is_rhel(version=""):
    return _check("Red Hat Enterprise Linux release {}".format(version))


def _check(text):
    with open("/etc/redhat-release") as f:
        line = f.readline()
    return text in line



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import logging

from .. import backends
from .. import cors
from .. import errors
from .. import http
from .. import ops
from .. import validate

log = logging.getLogger("images")

BASE_FEATURES = ("checksum", "extents")
ALL_FEATURES = BASE_FEATURES + ("flush", "zero")


class Handler:
    """
    Handle requests for the /images/ resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    @cors.allow()
    def put(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # Used by engine UI to close connection after upload.
        close = req.query.get("close") == "y"
        if close:
            resp.close_connection()

        size = req.content_length
        if size is None:
            raise http.Error(
                http.BAD_REQUEST, "Content-Length header is required")

        offset = req.content_range.first if req.content_range else 0

        # For backward compatibility, we flush by default.
        flush = validate.enum(req.query, "flush", ("y", "n"), default="y")
        flush = (flush == "y")

        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        validate.allowed_range(offset, size, ticket)

        log.debug(
            "[%s] WRITE size=%d offset=%d flush=%s close=%s transfer=%s",
            req.client_addr, size, offset, flush, close, ticket.transfer_id)

        op = ops.Write(
            ctx.backend,
            req,
            ctx.buffer,
            size,
            offset=offset,
            flush=flush,
            clock=req.clock)
        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    @cors.allow()
    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # Used by engine UI to close connection after download.
        close = req.query.get("close") == "y"
        if close:
            resp.close_connection()

        offset = 0
        size = None
        if req.range:
            offset = req.range.first
            if offset < 0:
                # TODO: support suffix-byte-range-spec "bytes=-last".
                # See https://tools.ietf.org/html/rfc7233#section-2.1.
                raise http.Error(
                    http.REQUESTED_RANGE_NOT_SATISFIABLE,
                    "suffix-byte-range-spec not supported yet")

            if req.range.last is not None:
                # Add 1 to include the last byte in the payload.
                size = req.range.last - offset + 1
                # TODO: validate size with actual image size.

        try:
            ticket = self.auth.authorize(ticket_id, "read")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        if size is not None:
            validate.allowed_range(offset, size, ticket)
            validate.available_range(offset, size, ticket, ctx.backend)
        else:
            size = min(ticket.size, ctx.backend.size()) - offset

        log.debug(
            "[%s] READ size=%d offset=%d close=%s transfer=%s",
            req.client_addr, size, offset, close, ticket.transfer_id)

        content_disposition = "attachment"
        if ticket.filename:
            content_disposition += "; filename=%s" % ticket.filename

        resp.headers["content-length"] = size
        resp.headers["content-type"] = "application/octet-stream"
        resp.headers["content-disposition"] = content_disposition

        if req.range:
            resp.status_code = http.PARTIAL_CONTENT
            resp.headers["content-range"] = "bytes %d-%d/%d" % (
                offset, offset + size - 1, ticket.size)

        op = ops.Read(
            ctx.backend,
            resp,
            ctx.buffer,
            size,
            offset=offset,
            clock=req.clock)
        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    def patch(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # TODO: Reject requests with too big payloads. We know the maximum size
        # of a payload based on the keys and the size of offset and size.
        try:
            msg = json.loads(req.read())
        except ValueError as e:
            raise http.Error(
                http.BAD_REQUEST, "Invalid JSON message {}" .format(e))

        op = validate.enum(msg, "op", ("zero", "flush"))
        if op == "zero":
            return self._zero(req, resp, ticket_id, msg)
        elif op == "flush":
            return self._flush(req, resp, ticket_id, msg)
        else:
            raise RuntimeError("Unreachable")

    def _zero(self, req, resp, ticket_id, msg):
        size = validate.integer(msg, "size", minval=0)
        offset = validate.integer(msg, "offset", minval=0, default=0)
        flush = validate.boolean(msg, "flush", default=False)

        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        validate.allowed_range(offset, size, ticket)

        log.debug(
            "[%s] ZERO size=%d offset=%d flush=%s transfer=%s",
            req.client_addr, size, offset, flush, ticket.transfer_id)

        op = ops.Zero(
            ctx.backend,
            size,
            offset=offset,
            flush=flush,
            clock=req.clock)

        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None
        except errors.PartialContent as e:
            raise http.Error(http.BAD_REQUEST, str(e))

    def _flush(self, req, resp, ticket_id, msg):
        try:
            ticket = self.auth.authorize(ticket_id, "write")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        log.info("[%s] FLUSH transfer=%s",
                 req.client_addr, ticket.transfer_id)

        op = ops.Flush(ctx.backend, clock=req.clock)

        try:
            ticket.run(op)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e)) from None

    @cors.allow(allow_methods="OPTIONS,GET,PUT")
    def options(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        options = {}

        if self.config.local.enable:
            options["unix_socket"] = self.config.local.socket

        if ticket_id == "*":
            log.info("[%s] OPTIONS transfer=*", req.client_addr)
            # Reporting the meta-capabilities for all images.
            allow = ["OPTIONS", "GET", "PUT", "PATCH"]
            options["features"] = ALL_FEATURES
        else:
            # Reporting real image capabilities per ticket.
            try:
                ticket = self.auth.authorize(ticket_id, "read")
                ctx = backends.get(req, ticket, self.config)
            except errors.AuthorizationError as e:
                resp.close_connection()
                raise http.Error(http.FORBIDDEN, str(e))

            log.info("[%s] OPTIONS transfer=%s",
                     req.client_addr, ticket.transfer_id)

            # Accessing ticket options considered as client activity.
            ticket.touch()

            allow = ["OPTIONS"]

            if ticket.may("read"):
                allow.append("GET")
                options["features"] = BASE_FEATURES

            if ticket.may("write"):
                allow.extend(("PUT", "PATCH"))
                options["features"] = ALL_FEATURES

            # Backend specific options.
            options["max_readers"] = ctx.backend.max_readers
            options["max_writers"] = ctx.backend.max_writers

        resp.headers["allow"] = ",".join(allow)
        resp.send_json(options)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import http.client as http_client

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server
from ovirt_imageio._internal.units import KiB

from . import testutil
from . import http


TICKET_SIZE = 16 * KiB


def test_start_with_ticket(tmpdir):
    # Create ticket
    image = testutil.create_tempfile(tmpdir, name="disk.raw", size=TICKET_SIZE)
    ticket = testutil.create_ticket(size=TICKET_SIZE, url=f'file://{image}')
    ticket_id = ticket['uuid']
    ticket_path = tmpdir.join("file.json")
    ticket_path.write(json.dumps(ticket))

    # Start server with ticket
    cfg = config.load("test/conf/daemon.conf")
    srv = server.Server(cfg, ticket=ticket_path)
    srv.start()
    try:
        data = b"a" * (TICKET_SIZE // 2) + b"b" * (TICKET_SIZE // 2)
        # Test that we can upload
        with http.RemoteClient(srv.config) as c:
            res = c.put(f'/images/{ticket_id}', data)
            assert res.status == http_client.OK
        # Test that we can download and matches the uploaded data
        with http.RemoteClient(srv.config) as c:
            res = c.get(f'/images/{ticket_id}')
            assert res.read() == data
            assert res.status == http_client.OK
    finally:
        srv.stop()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

from . import http


def enum(d, name, values, default=None):
    try:
        val = d[name]
    except KeyError:
        if default is not None:
            return default
        raise http.Error(
            http.BAD_REQUEST, "Missing required value for {}".format(name))

    if val not in values:
        raise http.Error(
            http.BAD_REQUEST,
            "Unsupported value {!r} for {!r}, expecting one of {}"
            .format(val, name, sorted(values)))

    return val


def integer(d, name, minval=None, maxval=None, default=None):
    try:
        val = d[name]
    except KeyError:
        if default is not None:
            return default
        raise http.Error(
            http.BAD_REQUEST, "Missing required value for {!r}".format(name))

    if not isinstance(val, int):
        raise http.Error(
            http.BAD_REQUEST, "Integer required {!r}".format(val))

    if minval is not None and val < minval:
        raise http.Error(
            http.BAD_REQUEST, "Invalid value {} < {}".format(val, minval))

    if maxval is not None and val > maxval:
        raise http.Error(
            http.BAD_REQUEST, "Invalid value {} > {}".format(val, maxval))

    return val


def boolean(d, name, default=False):
    try:
        val = d[name]
    except KeyError:
        if default is not None:
            return default
        raise http.Error(
            http.BAD_REQUEST, "Missing required value for {!r}".format(name))

    if not isinstance(val, bool):
        raise http.Error(
            http.BAD_REQUEST, "Boolean required {!r}".format(val))

    return val


def allowed_range(offset, size, ticket):
    """
    Checks that requested size is no greater than what's allowed by ticket,
    taking offset into account.
    """
    if offset + size > ticket.size:
        raise http.Error(http.REQUESTED_RANGE_NOT_SATISFIABLE,
                         "Requested range out of allowed range",
                         content_range="bytes */{}".format(ticket.size))


def available_range(offset, size, ticket, backend):
    """
    Checks that requested size is no greater than what's allowed by ticket and
    backend, taking offset into account.
    """
    requested = offset + size
    available = min(ticket.size, backend.size())
    if requested > available:
        raise http.Error(http.REQUESTED_RANGE_NOT_SATISFIABLE,
                         "Requested more bytes than available: {} > {}"
                         .format(requested, available),
                         content_range="bytes */{}".format(available))



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import logging
import subprocess

log = logging.getLogger("qemu_img")


class ContentMismatch(Exception):
    """
    Raised when compared images differ.
    """


class OpenImageError(Exception):
    """
    Raised when image cannot be opened, for example when specified format does
    not match the actual image format.
    """


def create(path, fmt, size=None, backing_file=None, backing_format=None,
           quiet=False):
    cmd = ["qemu-img", "create", "-f", fmt]
    if backing_file:
        cmd.extend(("-b", backing_file))
        if backing_format:
            cmd.extend(("-F", backing_format))
    if quiet:
        cmd.append("-q")
    cmd.append(path)
    if size is not None:
        cmd.append(str(size))
    subprocess.check_call(cmd)


def convert(src, dst, src_fmt, dst_fmt, progress=False, compressed=False):
    cmd = ["qemu-img", "convert", "-f", src_fmt, "-O", dst_fmt]
    if progress:
        cmd.append("-p")
    if compressed:
        cmd.append("-c")
    cmd.extend((src, dst))
    subprocess.check_call(cmd)


def unsafe_rebase(path, backing, backing_fmt):
    cmd = ["qemu-img", "rebase", "-u", "-b", backing, "-F", backing_fmt, path]
    subprocess.check_call(cmd)


def info(path):
    cmd = ["qemu-img", "info", "--output", "json", path]
    out = subprocess.check_output(cmd)
    return json.loads(out.decode("utf-8"))


def measure(path, out_fmt):
    cmd = ["qemu-img", "measure", "--output", "json", "-O", out_fmt, path]
    out = subprocess.check_output(cmd)
    return json.loads(out.decode("utf-8"))


def compare(a, b, format1=None, format2=None, strict=False):
    cmd = ["qemu-img", "compare"]

    if strict:
        cmd.append("-s")

    if format1:
        cmd.append("-f")
        cmd.append(format1)

    if format2:
        cmd.append("-F")
        cmd.append(format2)

    cmd.append(a)
    cmd.append(b)

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode == 0:
        return
    elif p.returncode == 1:
        raise ContentMismatch(out.decode("utf-8"))
    elif p.returncode == 2:
        raise OpenImageError(out.decode("utf-8"))
    else:
        raise RuntimeError(err.decode("utf-8"))


def bitmap_add(path, bitmap):
    cmd = ["qemu-img", "bitmap", "--add", path, bitmap]
    subprocess.check_call(cmd)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import logging
import socket
import time

from contextlib import closing

from . import ipv6

log = logging.getLogger("sockutil")


class UnixAddress(str):
    """
    A unix socket path with additioal methods to make it easier to handle both
    unix socket and TCP socket in the same code.

    Because we inherit from str, you can pass an instance to socket.connect()
    or socket.bind().
    """

    @property
    def transport(self):
        return "unix"

    @property
    def path(self):
        return str(self)


class TCPAddress(tuple):
    """
    A TCP socket 2 tuple (host, port) with additioal methods to make it easier
    to handle both unix socket and TCP socket in the same code.

    Because we inherit from tuple, you can pass an instance to socket.connect()
    or socket.bind().
    """

    def __new__(cls, host, port):
        if not isinstance(host, str):
            raise ValueError("Invalid host {!r}, expecting string value"
                             .format(host))
        if not isinstance(port, int):
            raise ValueError("Invalid port {!r}, expecting integer value"
                             .format(port))
        host = ipv6.unquote_address(host)
        return tuple.__new__(cls, (host, port))

    @property
    def transport(self):
        return "tcp"

    @property
    def host(self):
        return self[0]

    @property
    def port(self):
        return self[1]


def wait_for_socket(addr, timeout, step=0.02):
    """
    Wait until socket is available.

    Arguments:
        addr (sockutil.Address): server address
        timeout (double): time to wait for socket
        step (double): check internal

    Return True if socket is available, False if socket is not available within
    the requested timeout.
    """
    start = time.monotonic()
    deadline = start + timeout

    log.debug("Waiting for socket %s up to %.6f seconds", addr, timeout)

    while True:
        try:
            check_connection(addr)
        except socket.error as e:
            if e.args[0] not in (errno.ECONNREFUSED, errno.ENOENT):
                raise

            # Timed out?
            now = time.monotonic()
            if now >= deadline:
                return False

            # Wait until the next iteration, but not more than the
            # requested deadline.
            wait = min(step, deadline - now)
            time.sleep(wait)
        else:
            log.debug("Waited for %s %.6f seconds",
                      addr, time.monotonic() - start)
            return True


def check_connection(addr):
    if addr.transport == "unix":
        sock = socket.socket(socket.AF_UNIX)
        with closing(sock):
            sock.connect(addr)
    elif addr.transport == "tcp":
        sock = socket.create_connection(addr)
        sock.close()
    else:
        raise RuntimeError("Cannot wait for {}".format(addr))
